import numpy as _np
import numba as _numba
import multiprocessing as _multiprocessing
import concurrent.futures as _futures
from concurrent.futures import ProcessPoolExecutor as _ProcessPoolExecutor
from scipy.optimize import minimize_scalar as _minimize_scalar
from tqdm import tqdm as _tqdm
import yaml as _yaml
import matplotlib.pyplot as _plt
from . import lib as _lib


_plt.style.use("ggplot")


def nan_index(y: _np.ndarray) -> tuple[_np.ndarray, callable]:
    """Find indices of NaN values in an array."""

    return _np.isnan(y), lambda z: z.nonzero()[0]


def interpolate_nan(data: _np.ndarray) -> _np.ndarray:
    """Linear interpolattion of NaN values in an array `data`."""

    nans, x = nan_index(data)
    data[nans] = _np.interp(x(nans), x(~nans), data[~nans])
    return data


def calibrate_z(
    locs: _np.recarray, 
    info: list[dict], 
    d: float, 
    magnification_factor: float, 
    path: str | None = None,
) -> dict:
    """Given localizations of a calibration sample (e.g., gold beads at
    different z positions), calibrate the z-axis by fitting a polynomial
    to the mean spot width/height of each frame. See Huang et al.
    Science, 2008. DOI: 10.1126/science.1153529.
    
    Parameters
    ----------
    locs : _np.recarray
        Localizations of a calibration sample.
    info : list of dicts
        Information about the calibration sample, including the number 
        of frames.
    d : float
        Step size in nm, i.e., the distance between the z positions of
        the calibration sample.
    magnification_factor : float
        Magnification factor of the microscope, i.e., the ratio between
        the actual z position of the calibration sample and the 
        estimated z position from the localization data.
    path : str, optional
        Path to save the calibration data as a YAML file. If None, the
        calibration data will not be saved. Default is None.
    
    Returns
    -------
    calibration : dict
        Dictionary containing the calibration coefficients (i.e., 
        polynomial coefficients), number of frames, step size, and
        magnification factor.
    """

    n_frames = info[0]["Frames"]
    range = (n_frames - 1) * d
    frame_range = _np.arange(n_frames)
    z_range = -(frame_range * d - range / 2)  # negative so that the first frames of
    # a bottom-to-up scan are positive z coordinates.

    mean_sx = _np.array([_np.mean(locs.sx[locs.frame == _]) for _ in frame_range])
    mean_sy = _np.array([_np.mean(locs.sy[locs.frame == _]) for _ in frame_range])
    var_sx = _np.array([_np.var(locs.sx[locs.frame == _]) for _ in frame_range])
    var_sy = _np.array([_np.var(locs.sy[locs.frame == _]) for _ in frame_range])

    keep_x = (locs.sx - mean_sx[locs.frame]) ** 2 < var_sx[locs.frame]
    keep_y = (locs.sy - mean_sy[locs.frame]) ** 2 < var_sy[locs.frame]
    keep = keep_x & keep_y
    locs = locs[keep]

    # Fits calibration curve to the mean of each frame
    mean_sx = _np.array([_np.mean(locs.sx[locs.frame == _]) for _ in frame_range])
    mean_sy = _np.array([_np.mean(locs.sy[locs.frame == _]) for _ in frame_range])

    # Fix nan
    mean_sx = interpolate_nan(mean_sx)
    mean_sy = interpolate_nan(mean_sy)

    cx = _np.polyfit(z_range, mean_sx, 6, full=False)
    cy = _np.polyfit(z_range, mean_sy, 6, full=False)

    # Fits calibration curve to each localization
    # true_z = locs.frame * d - range / 2
    # cx = _np.polyfit(true_z, locs.sx, 6, full=False)
    # cy = _np.polyfit(true_z, locs.sy, 6, full=False)

    calibration = {
        "X Coefficients": [float(_) for _ in cx],
        "Y Coefficients": [float(_) for _ in cy],
        "Number of frames": int(n_frames),
        "Step size in nm": float(d),
        "Magnification factor": float(magnification_factor),
    }
    if path is not None:
        with open(path, "w") as f:
            _yaml.dump(calibration, f, default_flow_style=False)

    locs = fit_z(locs, info, calibration, magnification_factor)
    locs.z /= magnification_factor

    _plt.figure(figsize=(18, 10))

    _plt.subplot(231)
    # Plot this if calibration curve is fitted to each localization
    # _plt.plot(true_z, locs.sx, '.', label='x', alpha=0.2)
    # _plt.plot(true_z, locs.sy, '.', label='y', alpha=0.2)
    # _plt.plot(true_z, _np.polyval(cx, true_z), '0.3', lw=1.5, label='x fit')
    # _plt.plot(true_z, _np.polyval(cy, true_z), '0.3', lw=1.5, label='y fit')
    _plt.plot(z_range, mean_sx, ".-", label="x")
    _plt.plot(z_range, mean_sy, ".-", label="y")
    _plt.plot(z_range, _np.polyval(cx, z_range), "0.3", lw=1.5, label="x fit")
    _plt.plot(z_range, _np.polyval(cy, z_range), "0.3", lw=1.5, label="y fit")
    _plt.xlabel("Stage position")
    _plt.ylabel("Mean spot width/height")
    _plt.xlim(z_range.min(), z_range.max())
    _plt.legend(loc="best")

    ax = _plt.subplot(232)
    _plt.scatter(locs.sx, locs.sy, c="k", lw=0, alpha=0.1)
    _plt.plot(
        _np.polyval(cx, z_range),
        _np.polyval(cy, z_range),
        lw=1.5,
        label="calibration from fit of mean width/height",
    )
    _plt.plot()
    ax.set_aspect("equal")
    _plt.xlabel("Spot width")
    _plt.ylabel("Spot height")
    _plt.legend(loc="best")

    _plt.subplot(233)
    _plt.plot(locs.z, locs.sx, ".", label="x", alpha=0.2)
    _plt.plot(locs.z, locs.sy, ".", label="y", alpha=0.2)
    _plt.plot(z_range, _np.polyval(cx, z_range), "0.3", lw=1.5, label="calibration")
    _plt.plot(z_range, _np.polyval(cy, z_range), "0.3", lw=1.5)
    _plt.xlim(z_range.min(), z_range.max())
    _plt.xlabel("Estimated z")
    _plt.ylabel("Spot width/height")
    _plt.legend(loc="best")

    ax = _plt.subplot(234)
    _plt.plot(z_range[locs.frame], locs.z, ".k", alpha=0.1)
    _plt.plot(
        [z_range.min(), z_range.max()],
        [z_range.min(), z_range.max()],
        lw=1.5,
        label="identity",
    )
    _plt.xlim(z_range.min(), z_range.max())
    _plt.ylim(z_range.min(), z_range.max())
    ax.set_aspect("equal")
    _plt.xlabel("Stage position")
    _plt.ylabel("Estimated z")
    _plt.legend(loc="best")

    ax = _plt.subplot(235)
    deviation = locs.z - z_range[locs.frame]
    bins = _lib.calculate_optimal_bins(deviation, max_n_bins=1000)
    _plt.hist(deviation, bins)
    _plt.xlabel("Deviation to true position")
    _plt.ylabel("Occurence")

    ax = _plt.subplot(236)
    square_deviation = deviation**2
    mean_square_deviation_frame = [
        _np.mean(square_deviation[locs.frame == _]) for _ in frame_range
    ]
    rmsd_frame = _np.sqrt(mean_square_deviation_frame)
    _plt.plot(z_range, rmsd_frame, ".-", color="0.3")
    _plt.xlim(z_range.min(), z_range.max())
    _plt.gca().set_ylim(bottom=0)
    _plt.xlabel("Stage position")
    _plt.ylabel("Mean z precision")

    _plt.tight_layout(pad=2)

    if path is not None:
        dirname = path[0:-5]
        _plt.savefig(dirname + ".png", format="png", dpi=300)

    _plt.show()

    export = False
    # Export
    if export:
        print("Exporting...")
        _np.savetxt("mean_sx.txt", mean_sx, delimiter="/t")
        _np.savetxt("mean_sy.txt", mean_sy, delimiter="/t")
        _np.savetxt("locs_sx.txt", locs.sx, delimiter="/t")
        _np.savetxt("locs_sy.txt", locs.sy, delimiter="/t")
        _np.savetxt("cx.txt", cx, delimiter="/t")
        _np.savetxt("cy.txt", cy, delimiter="/t")
        _np.savetxt("z_range.txt", z_range, delimiter="/t")
        _np.savetxt("locs_z.txt", locs.z, delimiter="/t")
        _np.savetxt("z_range_locs_frame.txt", z_range[locs.frame], delimiter="/t")
        _np.savetxt("rmsd_frame.txt", rmsd_frame, delimiter="/t")

    # np.savetxt('test.out', x, delimiter=',')   # X is an array
    # np.savetxt('test.out', (x,y,z))   # x,y,z equal sized 1D arrays
    # np.savetxt('test.out', x, fmt='%1.4e')   # use exponential notation

    return calibration


@_numba.jit(nopython=True, nogil=True)
def _fit_z_target(
    z: _np.array, 
    sx: _np.array, 
    sy: _np.array, 
    cx: _np.array, 
    cy: _np.array,
) -> _np.array:
    """Target function that's to be minimized for fitting the z 
    coordinates given the single-emitter image width and height as well
    as the calibration curve coefficients. It calculates the difference 
    between the square root of the spot width/height and the polynomial 
    fit of the z-axis calibration curve."""

    z2 = z * z
    z3 = z * z2
    z4 = z * z3
    z5 = z * z4
    z6 = z * z5
    wx = (
        cx[0] * z6
        + cx[1] * z5
        + cx[2] * z4
        + cx[3] * z3
        + cx[4] * z2
        + cx[5] * z
        + cx[6]
    )
    wy = (
        cy[0] * z6
        + cy[1] * z5
        + cy[2] * z4
        + cy[3] * z3
        + cy[4] * z2
        + cy[5] * z
        + cy[6]
    )
    return (sx**0.5 - wx**0.5) ** 2 + (
        sy**0.5 - wy**0.5
    ) ** 2  # Apparently this results in slightly more accurate z coordinates
    # (Huang et al. '08)


def fit_z(
    locs: _np.recarray, 
    info: list[dict], 
    calibration: dict, 
    magnification_factor: float, 
    filter: int = 2
) -> _np.recarray:
    """Fits z coordinates to the localizations based on the calibration
    curve coefficients and the single-emitter image width and height.
    
    Parameters
    ----------
    locs : _np.recarray
        Localizations to fit the z-axis calibration curve to.
    info : list of dicts
        Information about the localizations, including the number of 
        frames.
    calibration : dict
        Calibration data containing the polynomial coefficients for
        the x and y axes, number of frames, step size, and magnification
        factor.
    magnification_factor : float
        Magnification factor of the microscope, i.e., the ratio between
        the actual z position of the calibration sample and the 
        estimated z position from the localization data.
    filter : int, optional
        Filter for the z fits. If set to 0, no filtering is applied.
        If set to 2, the z fits are filtered based on the root mean 
        square deviation (RMSD) of the z calibration. Default is 2.
    
    Returns
    -------
    locs : _np.recarray
        Localizations with the fitted z coordinates and their residuals
        (d_zcalib).
    """
    
    cx = _np.array(calibration["X Coefficients"])
    cy = _np.array(calibration["Y Coefficients"])
    z = _np.zeros_like(locs.x)
    square_d_zcalib = _np.zeros_like(z)
    sx = locs.sx
    sy = locs.sy
    for i in range(len(z)):
        result = _minimize_scalar(
            _fit_z_target, 
            bounds=[-1000, 1000], # to avoid potential gaps in the calibration curve, credits to Loek Andriessen
            args=(sx[i], sy[i], cx, cy)
        )
        z[i] = result.x
        square_d_zcalib[i] = result.fun
    z *= magnification_factor
    locs = _lib.append_to_rec(locs, z, "z")
    locs = _lib.append_to_rec(locs, _np.sqrt(square_d_zcalib), "d_zcalib")
    locs = _lib.ensure_sanity(locs, info)
    return filter_z_fits(locs, filter)


def fit_z_parallel(
    locs: _np.recarray, 
    info: list[dict], 
    calibration: dict, 
    magnification_factor: float, 
    filter: int = 2, 
    asynch: bool = False,
) -> _np.recarray | list[_futures.Future]:
    """Fits z coordinates to the localizations based on the calibration
    curve coefficients and the single-emitter image width and height,
    optionally using multiprocessing.
    
    Parameters
    ----------
    locs : _np.recarray
        Localizations to fit the z-axis calibration curve to.
    info : list of dicts
        Information about the localizations, including the number of 
        frames.
    calibration : dict
        Calibration data containing the polynomial coefficients for
        the x and y axes, number of frames, step size, and magnification
        factor.
    magnification_factor : float
        Magnification factor of the microscope, i.e., the ratio between
        the actual z position of the calibration sample and the 
        estimated z position from the localization data.
    filter : int, optional
        Filter for the z fits. If set to 0, no filtering is applied.
        If set to 2, the z fits are filtered based on the root mean
        square deviation (RMSD) of the z calibration. Default is 2.
    asynch : bool, optional
        If True, use multiprocessing. Then, a list of futures that can 
        be used to retrieve the results asynchronously is returned. If 
        False, the function waits for all tasks to complete and returns 
        the combined results. Default is False.
        
    Returns
    -------
    locs : _np.recarray or list of _futures.Future
        If `asynch` is False, returns a recarray of localizations with
        the fitted z coordinates and their residuals (d_zcalib).
        If `asynch` is True, returns a list of futures that can be
        used to retrieve the results asynchronously.
    """
    
    n_workers = min(
        60, max(1, int(0.75 * _multiprocessing.cpu_count()))
    ) # Python crashes when using >64 cores
    n_locs = len(locs)
    n_tasks = 100 * n_workers
    spots_per_task = [
        int(n_locs / n_tasks + 1) if _ < n_locs % n_tasks else int(n_locs / n_tasks)
        for _ in range(n_tasks)
    ]
    start_indices = _np.cumsum([0] + spots_per_task[:-1])
    fs = []
    executor = _ProcessPoolExecutor(n_workers)
    for i, n_locs_task in zip(start_indices, spots_per_task):
        fs.append(
            executor.submit(
                fit_z,
                locs[i : i + n_locs_task],
                info,
                calibration,
                magnification_factor,
                filter=0,
            )
        )
    if asynch:
        return fs
    with _tqdm(total=n_tasks, unit="task") as progress_bar:
        for f in _futures.as_completed(fs):
            progress_bar.update()
    return locs_from_futures(fs, filter=filter)


def locs_from_futures(
    futures: list[_futures.Future], 
    filter: int = 2
) -> _np.recarray:
    """Combines the results from a list of futures (i.e., 
    multiprocessing results) into a single recarray of localizations
    with fitted z coordinates and their residuals (d_zcalib).
    
    Parameters
    ----------
    futures : list of _futures.Future
        List of futures that contain the results of the z fits.
    filter : int, optional
        Filter for the z fits. If set to 0, no filtering is applied.
        If set to 2, the z fits are filtered based on the root mean
        square deviation (RMSD) of the z calibration. Default is 2. 
    
    Returns
    -------
    locs : _np.recarray
        Recarray of localizations with the fitted z coordinates and
        their residuals (d_zcalib).
    """

    locs = [_.result() for _ in futures]
    locs = _np.hstack(locs).view(_np.recarray)
    return filter_z_fits(locs, filter)


def filter_z_fits(locs: _np.recarray, range: int) -> _np.recarray:
    """Filters the z fits based on the root mean square deviation (RMSD)
    of the z calibration (d_zcalib residual). If `range` is set to 0, no 
    filtering is applied. If `range` is greater than 0, the 
    localizations with a RMSD greater than `range` are removed.

    Parameters
    ----------
    locs : _np.recarray
        Localizations with fitted z coordinates and their residuals
        (d_zcalib).
    range : int
        Range for filtering the z fits. If set to 0, no filtering is
        applied. If set to a positive value, localizations with a
        RMSD greater than `range` times the RMSD of the z calibration
        are removed.

    Returns
    -------
    locs : _np.recarray
        Recarray of localizations with the fitted z coordinates and
        their residuals (d_zcalib) after filtering.
    """

    if range > 0:
        rmsd = _np.sqrt(_np.nanmean(locs.d_zcalib**2))
        locs = locs[locs.d_zcalib <= range * rmsd]
    return locs



"""
Streamlit application to interface with the database
"""

import os
import socket
from PIL import Image
import streamlit as st
import pandas as pd
from sqlalchemy import create_engine
from datetime import datetime

from picasso import localize
from status import status
from preview import preview
from history import history
from watcher import watcher
from compare import compare

from picasso.__version__ import VERSION_NO

st.set_page_config(layout="wide")


_this_file = os.path.abspath(__file__)
_this_directory = os.path.dirname(_this_file)

LOGO_PATH = os.path.abspath(
    os.path.join(_this_directory, os.pardir, "gui/icons/picasso_server.png")
)
logo = Image.open(LOGO_PATH)

c1, c2, c3, c4 = st.sidebar.columns((1, 1, 1, 1))
c1.image(logo)
c2.write("# Picasso Server")

engine = create_engine("sqlite:///" + localize._db_filename(), echo=False)

st.sidebar.code(f"{socket.gethostname()}\nVersion {VERSION_NO}")

sidebar = {
    "Status": status,
    "History": history,
    "Compare": compare,
    "Watcher": watcher,
    "Preview": preview,
}

menu = st.sidebar.radio("", list(sidebar.keys()))

if menu:
    sidebar[menu]()



"""
    gui/average3
    ~~~~~~~~~~~~~~~~~~~~

    Graphical user interface for three-dimensional averaging of particles

    :author: Maximilian Strauss, 2017-2018
    :copyright: Copyright (c) 2017-2018 Jungmann Lab, MPI of Biochemistry
"""
import os.path
import sys
import traceback
import colorsys

import matplotlib.pyplot as plt
import numba
import numpy as np
import scipy
from scipy import signal

from PyQt5 import QtCore, QtGui, QtWidgets

from .. import io, lib, render, __version__

from numpy.lib.recfunctions import stack_arrays

from cmath import rect, phase

import scipy.ndimage.filters
import importlib, pkgutil

DEFAULT_OVERSAMPLING = 1.0
INITIAL_REL_MAXIMUM = 2.0
ZOOM = 10 / 7
N_GROUP_COLORS = 8


@numba.jit(nopython=True, nogil=True)
def render_hist(x, y, oversampling, t_min, t_max):
    n_pixel = int(np.ceil(oversampling * (t_max - t_min)))
    in_view = (x > t_min) & (y > t_min) & (x < t_max) & (y < t_max)
    x = x[in_view]
    y = y[in_view]
    x = oversampling * (x - t_min)
    y = oversampling * (y - t_min)
    image = np.zeros((n_pixel, n_pixel), dtype=np.float32)
    render._fill(image, x, y)
    return len(x), image


@numba.jit(nopython=True, nogil=True)
def render_histxyz(a, b, oversampling, a_min, a_max, b_min, b_max):
    n_pixel_a = int(np.ceil(oversampling * (a_max - a_min)))
    n_pixel_b = int(np.ceil(oversampling * (b_max - b_min)))
    in_view = (a > a_min) & (b > b_min) & (a < a_max) & (b < b_max)
    a = a[in_view]
    b = b[in_view]
    a = oversampling * (a - a_min)
    b = oversampling * (b - b_min)
    image = np.zeros((n_pixel_b, n_pixel_a), dtype=np.float32)
    render._fill(image, a, b)

    return len(a), image


def rotate_axis(axis, vx, vy, vz, angle, pixelsize):
    if axis == "z":
        vx_rot = np.cos(angle) * vx - np.sin(angle) * vy
        vy_rot = np.sin(angle) * vx + np.cos(angle) * vy
        vz_rot = vz
    elif axis == "y":
        vx_rot = np.cos(angle) * vx + np.sin(angle) * np.divide(vz, pixelsize)
        vy_rot = vy
        vz_rot = -np.sin(angle) * vx * pixelsize + np.cos(angle) * vz
    elif axis == "x":
        vx_rot = vx
        vy_rot = np.cos(angle) * vy - np.sin(angle) * np.divide(vz, pixelsize)
        vz_rot = np.sin(angle) * vy * pixelsize + np.cos(angle) * vz
    return vx_rot, vy_rot, vz_rot


def compute_xcorr(CF_image_avg, image):
    F_image = np.fft.fft2(image)
    xcorr = np.fft.fftshift(np.real(np.fft.ifft2((F_image * CF_image_avg))))
    return xcorr


class ParametersDialog(QtWidgets.QDialog):
    def __init__(self, window):
        super().__init__(window)
        self.window = window
        self.setWindowTitle("Parameters")
        self.setModal(False)
        grid = QtWidgets.QGridLayout(self)

        grid.addWidget(QtWidgets.QLabel("Oversampling:"), 0, 0)
        self.oversampling = QtWidgets.QDoubleSpinBox()
        self.oversampling.setRange(1, 200)
        self.oversampling.setValue(DEFAULT_OVERSAMPLING)
        self.oversampling.setDecimals(1)
        self.oversampling.setKeyboardTracking(False)
        self.oversampling.valueChanged.connect(self.window.updateLayout)
        grid.addWidget(self.oversampling, 0, 1)

        self.iterations = QtWidgets.QSpinBox()
        self.iterations.setRange(1, 1)
        self.iterations.setValue(1)


class View(QtWidgets.QLabel):
    def __init__(self, window):
        super().__init__()
        self.window = window
        self.setMinimumSize(1, 1)
        self.setAlignment(QtCore.Qt.AlignCenter)
        self.setAcceptDrops(True)
        self._pixmap = None

    def dragEnterEvent(self, event):
        if event.mimeData().hasUrls():
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        urls = event.mimeData().urls()
        path = urls[0].toLocalFile()
        ext = os.path.splitext(path)[1].lower()
        if ext == ".hdf5":
            self.open(path)

    def resizeEvent(self, event):
        if self._pixmap is not None:
            self.set_pixmap(self._pixmap)

    def set_image(self, image):
        cmap = np.uint8(np.round(255 * plt.get_cmap("magma")(np.arange(256))))
        image /= image.max()
        image = np.minimum(image, 1.0)
        image = np.round(255 * image).astype("uint8")
        Y, X = image.shape
        self._bgra = np.zeros((Y, X, 4), dtype=np.uint8, order="C")
        self._bgra[..., 0] = cmap[:, 2][image]
        self._bgra[..., 1] = cmap[:, 1][image]
        self._bgra[..., 2] = cmap[:, 0][image]
        qimage = QtGui.QImage(self._bgra.data, X, Y, QtGui.QImage.Format_RGB32)
        self._pixmap = QtGui.QPixmap.fromImage(qimage)
        self.set_pixmap(self._pixmap)

    def set_pixmap(self, pixmap):
        self.setPixmap(
            pixmap.scaled(
                self.width(),
                self.height(),
                QtCore.Qt.KeepAspectRatio,
                QtCore.Qt.FastTransformation,
            )
        )

    def update_image(self, *args):
        oversampling = self.window.parameters_dialog.oversampling.value()
        t_min = -self.r
        t_max = self.r
        N_avg, image_avg = render.render_hist(
            self.locs, oversampling, t_min, t_min, t_max, t_max
        )
        self.set_image(image_avg)


class DatasetDialog(QtWidgets.QDialog):
    def __init__(self, window):
        super().__init__(window)
        self.window = window
        self.setWindowTitle("Datasets")
        self.setModal(False)
        self.layout = QtWidgets.QVBoxLayout()
        self.checks = []
        self.setLayout(self.layout)

    def add_entry(self, path):
        c = QtWidgets.QCheckBox(path)
        self.layout.addWidget(c)
        self.checks.append(c)
        self.checks[-1].setChecked(True)


class Window(QtWidgets.QMainWindow):
    def __init__(self):
        super().__init__()
        self.setWindowTitle(f"Picasso v{__version__}: Average3")
        self.resize(1024, 512)
        this_directory = os.path.dirname(os.path.realpath(__file__))
        icon_path = os.path.join(this_directory, "icons", "average.ico")
        icon = QtGui.QIcon(icon_path)
        self.setWindowIcon(icon)
        self.setAcceptDrops(True)
        self.parameters_dialog = ParametersDialog(self)
        self.dataset_dialog = DatasetDialog(self)
        menu_bar = self.menuBar()
        file_menu = menu_bar.addMenu("File")
        open_action = file_menu.addAction("Open")
        open_action.setShortcut(QtGui.QKeySequence.Open)
        open_action.triggered.connect(self.open)
        file_menu.addAction(open_action)
        save_action = file_menu.addAction("Save")
        save_action.setShortcut(QtGui.QKeySequence.Save)
        save_action.triggered.connect(self.save)
        file_menu.addAction(save_action)
        process_menu = menu_bar.addMenu("Process")
        parameters_action = process_menu.addAction("Parameters")
        parameters_action.setShortcut("Ctrl+P")
        parameters_action.triggered.connect(self.parameters_dialog.show)
        dataset_action = process_menu.addAction("Datasets")
        dataset_action.triggered.connect(self.dataset_dialog.show)

        self.status_bar = self.statusBar()
        self._pixmap = None
        self.locs = []
        self.z_state = []
        self.group_index = []
        self.infos = []
        self.locs_paths = []
        self._mode = "Zoom"
        self._pan = False
        self._size_hint = (768, 768)
        self.n_locs = 0
        self._picks = []
        self.index_blocks = []
        self._drift = []

        # Define DisplaySettingsDialog
        self.viewxy = QtWidgets.QLabel("")
        self.viewxz = QtWidgets.QLabel("")
        self.viewyz = QtWidgets.QLabel("")
        self.viewcp = QtWidgets.QLabel("")

        minsize = 512
        self.viewxy.setFixedWidth(minsize)
        self.viewxy.setFixedHeight(minsize)
        self.viewxz.setFixedWidth(minsize)
        self.viewxz.setFixedHeight(minsize)
        self.viewyz.setFixedWidth(minsize)
        self.viewyz.setFixedHeight(minsize)
        self.viewcp.setFixedWidth(minsize)
        self.viewcp.setFixedHeight(minsize)

        # Define layout
        display_groupbox = QtWidgets.QGroupBox("Display")
        displaygrid = QtWidgets.QGridLayout(display_groupbox)
        displaygrid.addWidget(QtWidgets.QLabel("XY"), 0, 0)
        displaygrid.addWidget(self.viewxy, 1, 0)
        displaygrid.addWidget(QtWidgets.QLabel("XZ"), 0, 1)
        displaygrid.addWidget(self.viewxz, 1, 1)
        displaygrid.addWidget(QtWidgets.QLabel("YZ"), 2, 0)
        displaygrid.addWidget(self.viewyz, 3, 0)
        displaygrid.addWidget(QtWidgets.QLabel("CP"), 2, 1)
        displaygrid.addWidget(self.viewcp, 3, 1)

        button_groupbox = QtWidgets.QGroupBox("Buttons")
        buttongrid = QtWidgets.QGridLayout(button_groupbox)

        rotation_groupbox = QtWidgets.QGroupBox("Rotation + Translation")
        rotationgrid = QtWidgets.QGridLayout(rotation_groupbox)
        centerofmassbtn = QtWidgets.QPushButton("Center of Mass XYZ")

        axis_groupbox = QtWidgets.QGroupBox("Axis")
        axisgrid = QtWidgets.QGridLayout(axis_groupbox)

        self.x_axisbtn = QtWidgets.QRadioButton("X")
        self.y_axisbtn = QtWidgets.QRadioButton("Y")
        self.z_axisbtn = QtWidgets.QRadioButton("Z")

        self.z_axisbtn.setChecked(True)

        axisgrid.addWidget(self.x_axisbtn, 0, 0)
        axisgrid.addWidget(self.y_axisbtn, 0, 1)
        axisgrid.addWidget(self.z_axisbtn, 0, 2)

        proj_groupbox = QtWidgets.QGroupBox("Projection")
        projgrid = QtWidgets.QGridLayout(proj_groupbox)

        self.xy_projbtn = QtWidgets.QRadioButton("XY")
        self.yz_projbtn = QtWidgets.QRadioButton("YZ")
        self.xz_projbtn = QtWidgets.QRadioButton("XZ")

        self.xy_projbtn.setChecked(True)

        projgrid.addWidget(self.xy_projbtn, 0, 0)
        projgrid.addWidget(self.yz_projbtn, 0, 1)
        projgrid.addWidget(self.xz_projbtn, 0, 2)

        rotatebtn = QtWidgets.QPushButton("Rotate")

        self.radio_sym = QtWidgets.QRadioButton("x symmetry")
        self.symEdit = QtWidgets.QSpinBox()
        self.symEdit.setRange(2, 100)
        self.symEdit.setValue(8)

        self.radio_sym_custom = QtWidgets.QRadioButton("custom symmetry")
        self.symcustomEdit = QtWidgets.QLineEdit("90,180,270")

        deg_groupbox = QtWidgets.QGroupBox("Degrees")
        deggrid = QtWidgets.QGridLayout(deg_groupbox)

        self.full_degbtn = QtWidgets.QRadioButton("Full")
        self.part_degbtn = QtWidgets.QRadioButton("Part")
        self.degEdit = QtWidgets.QTextEdit()

        self.degEdit = QtWidgets.QSpinBox()
        self.degEdit.setRange(1, 10)
        self.degEdit.setValue(5)

        deggrid.addWidget(self.full_degbtn, 0, 0)
        deggrid.addWidget(self.part_degbtn, 0, 1)
        deggrid.addWidget(self.degEdit, 0, 2)

        self.full_degbtn.setChecked(True)

        # Rotation Groupbox
        rotationgrid.addWidget(axis_groupbox, 0, 0, 1, 2)
        rotationgrid.addWidget(proj_groupbox, 1, 0, 1, 2)
        rotationgrid.addWidget(deg_groupbox, 2, 0, 1, 2)
        rotationgrid.addWidget(rotatebtn, 3, 0, 1, 2)
        rotationgrid.addWidget(self.symEdit, 4, 0)
        rotationgrid.addWidget(self.radio_sym, 4, 1)
        rotationgrid.addWidget(self.radio_sym_custom, 5, 0)
        rotationgrid.addWidget(self.symcustomEdit, 5, 1)

        buttongrid.addWidget(centerofmassbtn, 0, 0)
        buttongrid.addWidget(rotation_groupbox, 1, 0)

        centerofmassbtn.clicked.connect(self.centerofmass)
        rotatebtn.clicked.connect(self.rotate_groups)

        self.translatebtn = QtWidgets.QCheckBox("Translate only")

        self.flipbtn = QtWidgets.QCheckBox("Consider flipped structures")

        self.alignxbtn = QtWidgets.QPushButton("Align X")
        self.alignybtn = QtWidgets.QPushButton("Align Y")
        self.alignzzbtn = QtWidgets.QPushButton("Align Z_Z")
        self.alignzybtn = QtWidgets.QPushButton("Align Z_Y")

        self.translatexbtn = QtWidgets.QPushButton("Translate X")
        self.translateybtn = QtWidgets.QPushButton("Translate Y")
        self.translatezbtn = QtWidgets.QPushButton("Translate Z")

        self.rotatexy_convbtn = QtWidgets.QPushButton("Rotate XY - Convolution")

        self.scorebtn = QtWidgets.QPushButton("Calculate Score")

        operate_groupbox = QtWidgets.QGroupBox("Operate")
        operategrid = QtWidgets.QGridLayout(operate_groupbox)

        rotationgrid.addWidget(self.translatebtn, 7, 0)

        rotationgrid.addWidget(self.flipbtn, 8, 0)

        self.x_range = QtWidgets.QLineEdit("-3,3")
        rotationgrid.addWidget(QtWidgets.QLabel("x-Range (Px)"), 9, 0)
        rotationgrid.addWidget(self.x_range, 9, 1)

        self.y_range = QtWidgets.QLineEdit("-3,3")
        rotationgrid.addWidget(QtWidgets.QLabel("y-Range (Px)"), 10, 0)
        rotationgrid.addWidget(self.y_range, 10, 1)

        self.z_range = QtWidgets.QLineEdit("-1000,1000")

        rotationgrid.addWidget(QtWidgets.QLabel("z-Range (nm)"), 11, 0)
        rotationgrid.addWidget(self.z_range, 11, 1)

        self.z_range.textChanged.connect(self.adjust_z)

        self.x_range.textChanged.connect(self.adjust_xy)

        self.y_range.textChanged.connect(self.adjust_xy)

        operategrid.addWidget(self.alignxbtn, 0, 1)
        operategrid.addWidget(self.alignybtn, 1, 1)
        operategrid.addWidget(self.alignzzbtn, 2, 1)
        operategrid.addWidget(self.alignzybtn, 3, 1)
        operategrid.addWidget(self.translatexbtn, 0, 0)
        operategrid.addWidget(self.translateybtn, 1, 0)
        operategrid.addWidget(self.translatezbtn, 2, 0)
        operategrid.addWidget(self.rotatexy_convbtn, 4, 0)
        operategrid.addWidget(self.scorebtn, 4, 1)

        self.rotatexy_convbtn.clicked.connect(self.rotatexy_convolution)

        self.alignxbtn.clicked.connect(self.align_x)
        self.alignybtn.clicked.connect(self.align_y)
        self.alignzzbtn.clicked.connect(self.align_zz)
        self.alignzybtn.clicked.connect(self.align_zy)

        self.translatexbtn.clicked.connect(self.translate_x)
        self.translateybtn.clicked.connect(self.translate_y)
        self.translatezbtn.clicked.connect(self.translate_z)

        self.scorebtn.clicked.connect(self.calculate_score)

        buttongrid.addWidget(operate_groupbox, 2, 0)

        self.contrastEdit = QtWidgets.QDoubleSpinBox()
        self.contrastEdit.setDecimals(1)
        self.contrastEdit.setRange(0, 10)
        self.contrastEdit.setValue(0.5)
        self.contrastEdit.setSingleStep(0.1)

        self.contrastEdit.valueChanged.connect(self.updateLayout)

        self.grid = QtWidgets.QGridLayout()
        self.grid.addWidget(display_groupbox, 0, 0, 2, 1)
        self.grid.addWidget(button_groupbox, 0, 1, 1, 1)

        contrast_groupbox = QtWidgets.QGroupBox("Contrast")
        contrastgrid = QtWidgets.QGridLayout(contrast_groupbox)

        contrastgrid.addWidget(self.contrastEdit)
        buttongrid.addWidget(contrast_groupbox)

        MODEL_X_DEFAULT = "0,20,40,60,0,20,40,60,0,20,40,60"
        MODEL_Y_DEFAULT = "0,20,40,0,20,40,0,20,40,0,20,40"
        MODEL_Z_DEFAULT = "0,0,0,0,0,0,0,0,0,0,0,0"

        self.modelchk = QtWidgets.QCheckBox("Use Model")
        self.model_x = QtWidgets.QLineEdit(MODEL_X_DEFAULT)
        self.model_y = QtWidgets.QLineEdit(MODEL_Y_DEFAULT)
        self.model_z = QtWidgets.QLineEdit(MODEL_Z_DEFAULT)

        self.model_preview_btn = QtWidgets.QPushButton("Preview")

        self.model_preview_btn.clicked.connect(self.model_preview)

        self.modelblurEdit = QtWidgets.QDoubleSpinBox()
        self.modelblurEdit.setDecimals(1)
        self.modelblurEdit.setRange(0, 10)
        self.modelblurEdit.setValue(0.5)
        self.modelblurEdit.setSingleStep(0.1)

        self.pixelsizeEdit = QtWidgets.QSpinBox()
        self.pixelsizeEdit.setRange(1, 999)
        self.pixelsizeEdit.setValue(130)

        model_groupbox = QtWidgets.QGroupBox("Model")
        modelgrid = QtWidgets.QGridLayout(model_groupbox)

        modelgrid.addWidget(self.modelchk, 0, 0)
        modelgrid.addWidget(QtWidgets.QLabel("X-Coordinates"), 1, 0)
        modelgrid.addWidget(self.model_x, 1, 1)
        modelgrid.addWidget(QtWidgets.QLabel("Y-Coordinates"), 2, 0)
        modelgrid.addWidget(self.model_y, 2, 1)
        modelgrid.addWidget(QtWidgets.QLabel("Z-Coordinates"), 3, 0)
        modelgrid.addWidget(self.model_z, 3, 1)
        modelgrid.addWidget(QtWidgets.QLabel("Blur:"), 4, 0)
        modelgrid.addWidget(self.modelblurEdit, 4, 1)
        modelgrid.addWidget(QtWidgets.QLabel("Pixelsize:"), 5, 0)
        modelgrid.addWidget(self.pixelsizeEdit, 5, 1)
        modelgrid.addWidget(self.model_preview_btn, 6, 0)
        modelgrid.addWidget(self.modelchk, 6, 1)

        buttongrid.addWidget(model_groupbox)

        mainWidget = QtWidgets.QWidget()
        mainWidget.setLayout(self.grid)
        self.setCentralWidget(mainWidget)
        self.status_bar.showMessage("Average3 ready.")

    def open(self):
        path, exe = QtWidgets.QFileDialog.getOpenFileName(
            self, "Open localizations", filter="*.hdf5"
        )
        if path:
            self.add(path)

    def save(self, path):
        n_channels = len(self.locs)
        for i in range(n_channels):
            cx = self.infos[i][0]["Width"] / 2
            cy = self.infos[i][0]["Height"] / 2
            out_locs = self.locs[i].copy()
            out_locs.x += cx
            out_locs.y += cy
            info = self.infos[i] + [{"Generated by": f"Picasso v{__version__} Average3"}]

            if not self.z_state[i]:
                out_locs = lib.remove_from_rec(out_locs, "z")

            out_path = os.path.splitext(self.locs_paths[i])[0] + "_avg3.hdf5"
            path, exe = QtWidgets.QFileDialog.getSaveFileName(
                self, "Save localizations", out_path, filter="*.hdf5"
            )
            io.save_locs(path, out_locs, info)

    def dragEnterEvent(self, event):
        if event.mimeData().hasUrls():
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        urls = event.mimeData().urls()
        path = urls[0].toLocalFile()
        ext = os.path.splitext(path)[1].lower()
        if ext == ".hdf5":
            print("Opening {} ..".format(path))
            self.add(path)

    def add(self, path, rendermode=True):
        try:
            locs, info = io.load_locs(path, qt_parent=self)
        except io.NoMetadataFileError:
            return

        if len(self.locs) == 0:
            self.pixelsize = 0

        if not hasattr(locs, "group"):
            msgBox = QtWidgets.QMessageBox(self)
            msgBox.setWindowTitle("Error")
            msgBox.setText(
                (
                    "Datafile does not contain group information."
                    " Please load file with picked localizations."
                )
            )
            msgBox.exec_()

        else:
            locs = lib.ensure_sanity(locs, info)
            if not hasattr(locs, "z"):
                locs = lib.append_to_rec(locs, locs.x.copy(), "z")
                self.pixelsize = 1
                has_z = False

            else:
                has_z = True
                if self.pixelsize == 0:
                    pixelsize, ok = QtWidgets.QInputDialog.getInt(
                        self,
                        "Pixelsize Dialog",
                        "Please enter the pixelsize in nm",
                        130,
                    )
                    if ok:
                        self.pixelsize = pixelsize
                    else:
                        self.pixelsize = 130

            self.locs.append(locs)
            self.z_state.append(has_z)
            self.infos.append(info)
            self.locs_paths.append(path)
            self.index_blocks.append(None)
            self._drift.append(None)
            self.dataset_dialog.add_entry(path)
            self.dataset_dialog.checks[-1].stateChanged.connect(self.updateLayout)

            cx = self.infos[-1][0]["Width"] / 2
            cy = self.infos[-1][0]["Height"] / 2

            self.locs[-1].x -= cx
            self.locs[-1].y -= cy

            if len(self.locs) == 1:
                self.median_lp = np.mean([np.median(locs.lpx), np.median(locs.lpy)])
                if hasattr(locs, "group"):
                    groups = np.unique(locs.group)
                    groupcopy = locs.group.copy()
                    for i in range(len(groups)):
                        groupcopy[locs.group == groups[i]] = i
                    np.random.shuffle(groups)
                    groups %= N_GROUP_COLORS
                    self.group_color = groups[groupcopy]
                if render:
                    self.fit_in_view(autoscale=True)
            else:
                if render:
                    self.update_scene()

            self.oversampling = 1
            if len(self.locs) == 1:
                self.t_min = np.min([np.min(locs.x), np.min(locs.y)])
                self.t_max = np.max([np.max(locs.x), np.max(locs.y)])
                self.z_min = np.min(locs.z)
                self.z_max = np.max(locs.z)
            else:
                self.t_min = np.min([np.min(locs.x), np.min(locs.y), self.t_min])
                self.t_max = np.max([np.max(locs.x), np.max(locs.y), self.t_max])
                self.z_min = np.min([np.min(locs.z), self.z_min])
                self.z_max = np.min([np.max(locs.z), self.z_max])

            if len(self.locs) == 1:
                print("Dataset loaded from {}.".format(path))
            else:
                print(
                    ("Dataset loaded from {}," " Total number of datasets {}.").format(
                        path, len(self.locs)
                    )
                )

            # CREATE GROUP INDEX
            if hasattr(locs, "group"):
                groups = np.unique(locs.group)
                n_groups = len(groups)
                n_locs = len(locs)

                group_index = scipy.sparse.lil_matrix((n_groups, n_locs), dtype=bool)
                progress = lib.ProgressDialog(
                    "Creating group index", 0, len(groups), self
                )
                progress.set_value(0)
                for i, group in enumerate(groups):
                    index = np.where(locs.group == group)[0]
                    group_index[i, index] = True
                    progress.set_value(i + 1)

                self.group_index.append(group_index)
                self.n_groups = n_groups

            os.chdir(os.path.dirname(path))

            self.calculate_radii()
            self.oversampling = 4
            self.updateLayout()

    def updateLayout(self):
        if len(self.locs) > 0:
            pixmap1, pixmap2, pixmap3 = self.hist_multi_channel(self.locs)
            self.viewxy.setPixmap(pixmap1)
            self.viewxz.setPixmap(pixmap2)
            self.viewyz.setPixmap(pixmap3)

    def centerofmass_all(self):
        # Align all by center of mass
        n_channels = len(self.locs)

        out_locs_x = []
        out_locs_y = []
        out_locs_z = []
        for j in range(n_channels):
            sel_locs_x = []
            sel_locs_y = []
            sel_locs_z = []

            # stack arrays
            sel_locs_x = self.locs[j].x
            sel_locs_y = self.locs[j].y
            sel_locs_z = self.locs[j].z
            out_locs_x.append(sel_locs_x)
            out_locs_y.append(sel_locs_y)
            out_locs_z.append(sel_locs_z)

        out_locs_x = stack_arrays(out_locs_x, asrecarray=True, usemask=False)
        out_locs_y = stack_arrays(out_locs_y, asrecarray=True, usemask=False)
        out_locs_z = stack_arrays(out_locs_z, asrecarray=True, usemask=False)

        mean_x = np.mean(out_locs_x)
        mean_y = np.mean(out_locs_y)
        mean_z = np.mean(out_locs_z)

        for j in range(n_channels):
            self.locs[j].x -= mean_x
            self.locs[j].y -= mean_y
            self.locs[j].z -= mean_z

    def calculate_radii(self):
        # CALCULATE PROPER R VALUES
        n_channels = len(self.locs)
        self.r = 0
        self.r_z = 0
        for j in range(n_channels):
            self.r = np.max(
                [
                    3 * np.sqrt(np.mean(self.locs[j].x ** 2 + self.locs[j].y ** 2)),
                    self.r,
                ]
            )
            self.r_z = np.max([5 * np.sqrt(np.mean(self.locs[j].z ** 2)), self.r_z])
        self.t_min = -self.r
        self.t_max = self.r
        self.z_min = -self.r_z
        self.z_max = self.r_z

        self.z_min_load = self.z_min.copy()
        self.z_max_load = self.z_max.copy()

    def centerofmass(self):
        print("Aligning by center of mass.. ", end="", flush=True)
        n_groups = self.n_groups
        n_channels = len(self.locs)
        progress = lib.ProgressDialog("Aligning by center of mass", 0, n_groups, self)
        progress.set_value(0)

        for i in range(n_groups):
            out_locs_x = []
            out_locs_y = []
            out_locs_z = []
            for j in range(n_channels):
                sel_locs_x = []
                sel_locs_y = []
                sel_locs_z = []
                index = self.group_index[j][i, :].nonzero()[1]
                # stack arrays
                sel_locs_x = self.locs[j].x[index]
                sel_locs_y = self.locs[j].y[index]
                sel_locs_z = self.locs[j].z[index]

                out_locs_x.append(sel_locs_x)
                out_locs_y.append(sel_locs_y)
                out_locs_z.append(sel_locs_z)
                progress.set_value(i + 1)

            out_locs_x = stack_arrays(out_locs_x, asrecarray=True, usemask=False)
            out_locs_y = stack_arrays(out_locs_y, asrecarray=True, usemask=False)
            out_locs_z = stack_arrays(out_locs_z, asrecarray=True, usemask=False)

            mean_x = np.mean(out_locs_x)
            mean_y = np.mean(out_locs_y)
            mean_z = np.mean(out_locs_z)

            for j in range(n_channels):
                index = self.group_index[j][i, :].nonzero()[1]
                self.locs[j].x[index] -= mean_x
                self.locs[j].y[index] -= mean_y
                self.locs[j].z[index] -= mean_z

        self.calculate_radii()
        self.updateLayout()

        print("Complete.")

    def histtoImage(self, image):
        cmap = np.uint8(np.round(255 * plt.get_cmap("magma")(np.arange(256))))
        image /= image.max()
        image = np.minimum(image, 1.0)
        image = np.round(255 * image).astype("uint8")
        Y, X = image.shape
        self._bgra = np.zeros((Y, X, 4), dtype=np.uint8, order="C")
        self._bgra[..., 0] = cmap[:, 2][image]
        self._bgra[..., 1] = cmap[:, 1][image]
        self._bgra[..., 2] = cmap[:, 0][image]
        qimage = QtGui.QImage(self._bgra.data, X, Y, QtGui.QImage.Format_RGB32)

        qimage = qimage.scaled(
            self.viewxy.width(),
            int(np.round(self.viewxy.height() * Y / X)),
            QtCore.Qt.KeepAspectRatioByExpanding,
        )
        pixmap = QtGui.QPixmap.fromImage(qimage)

        return pixmap

    def hist_multi_channel(self, locs):

        oversampling = self.parameters_dialog.oversampling.value()
        self.oversampling = oversampling
        if locs is None:
            locs = self.locs
        n_channels = len(locs)

        hues = np.arange(0, 1, 1 / n_channels)
        colors = [colorsys.hsv_to_rgb(_, 1, 1) for _ in hues]

        renderings = []
        for i in range(n_channels):
            if self.dataset_dialog.checks[i].isChecked():
                renderings.append(
                    render.render_hist3d(
                        locs[i],
                        oversampling,
                        self.t_min,
                        self.t_min,
                        self.t_max,
                        self.t_max,
                        self.z_min,
                        self.z_max,
                        self.pixelsize,
                    )
                )
        images = np.array([_[1] for _ in renderings])

        pixmap1 = self.pixmap_from_colors(images, colors, 2)
        pixmap2 = self.pixmap_from_colors(images, colors, 0)
        pixmap3 = self.pixmap_from_colors(images, colors, 1)

        return pixmap1, pixmap2, pixmap3

    def pixmap_from_colors(self, images, colors, axisval):
        if axisval == 2:
            image = [np.sum(_, axis=axisval) for _ in images]
        else:
            image = [np.transpose(np.sum(_, axis=axisval)) for _ in images]
        image = np.array([self.scale_contrast(_) for _ in image])
        Y, X = image.shape[1:]
        bgra = np.zeros((Y, X, 4), dtype=np.float32)
        for color, image in zip(colors, image):
            bgra[:, :, 0] += color[2] * image
            bgra[:, :, 1] += color[1] * image
            bgra[:, :, 2] += color[0] * image
        bgra = np.minimum(bgra, 1)
        self._bgra = self.to_8bit(bgra)

        qimage = QtGui.QImage(self._bgra.data, X, Y, QtGui.QImage.Format_RGB32)

        qimage = qimage.scaled(
            self.viewxy.width(),
            int(np.round(self.viewxy.height() * Y / X)),
            QtCore.Qt.KeepAspectRatioByExpanding,
        )
        pixmap = QtGui.QPixmap.fromImage(qimage)

        return pixmap

    def align_x(self):
        print("Align X")
        self.align_all("x")

    def align_y(self):
        print("Align Y")
        self.align_all("y")

    def align_zz(self):
        print("Align Z")
        self.align_all("zz")

    def align_zy(self):
        print("Align Z")
        self.align_all("zy")

    def translate_x(self):
        print("Translate X")
        self.translate("x")

    def translate_y(self):
        print("Translate Y")
        self.translate("y")

    def translate_z(self):
        print("Translate Z")
        self.translate("z")

    def translate(self, translateaxis):
        renderings = [
            render.render_hist3d(
                _,
                self.oversampling,
                self.t_min,
                self.t_min,
                self.t_max,
                self.t_max,
                self.z_min,
                self.z_max,
                self.pixelsize,
            )
            for _ in self.locs
        ]

        images = np.array([_[1] for _ in renderings])

        if translateaxis == "x":
            image = [np.sum(_, axis=2) for _ in images]
            signalimg = [np.sum(_, axis=0) for _ in image]
        elif translateaxis == "y":
            image = [np.sum(_, axis=2) for _ in images]
            signalimg = [np.sum(_, axis=1) for _ in image]
        elif translateaxis == "z":
            image = [np.sum(_, axis=1) for _ in images]
            signalimg = [np.sum(_, axis=0) for _ in image]

        fig = plt.figure(figsize=(5, 5))
        ax1 = fig.add_subplot(1, 1, 1)
        for element in signalimg:
            plt.plot(element)
        n_groups = self.group_index[0].shape[0]
        print("Translating..")
        for i in range(n_groups):
            self.status_bar.showMessage("Group {} / {}.".format(i, n_groups))
            self.translate_group(signalimg, i, translateaxis)

        fig.canvas.draw()
        size = fig.canvas.size()
        width, height = size.width(), size.height()
        im = QtGui.QImage(
            fig.canvas.buffer_rgba(), width, height, QtGui.QImage.Format_ARGB32
        )
        self.viewcp.setPixmap((QtGui.QPixmap(im)))
        self.viewcp.setAlignment(QtCore.Qt.AlignCenter)
        plt.close(fig)

        self.centerofmass_all()
        self.updateLayout()
        self.status_bar.showMessage("Done!")

    def translate_group(self, signalimg, group, translateaxis):
        n_channels = len(self.locs)
        all_xcorr = np.zeros((1, n_channels))
        all_da = np.zeros((1, n_channels))

        if translateaxis == "x":
            proplane = "xy"
        elif translateaxis == "y":
            proplane = "xy"
        elif translateaxis == "z":
            proplane = "xz"

        plotmode = 0

        for j in range(n_channels):
            if plotmode:
                fig = plt.figure()
                ax1 = fig.add_subplot(1, 3, 1)
                plt.plot(signalimg[j])
                ax2 = fig.add_subplot(1, 3, 2)

            if self.dataset_dialog.checks[j].isChecked():
                index = self.group_index[j][group].nonzero()[1]
                x_rot = self.locs[j].x[index]
                y_rot = self.locs[j].y[index]
                z_rot = self.locs[j].z[index]

                plane = self.render_planes(
                    x_rot, y_rot, z_rot, proplane, self.pixelsize
                )  #
                if translateaxis == "x":
                    projection = np.sum(plane, axis=0)
                elif translateaxis == "y":
                    projection = np.sum(plane, axis=1)
                elif translateaxis == "z":
                    projection = np.sum(plane, axis=1)

                if plotmode:
                    plt.plot(projection)
                # print('Step X')
                # ax3 = fig.add_subplot(1,3,3)
                # plt.imshow(plane, interpolation='nearest', cmap=plt.cm.ocean)
                corrval = np.max(signal.correlate(signalimg[j], projection))
                shiftval = (
                    np.argmax(signal.correlate(signalimg[j], projection))
                    - len(signalimg[j])
                    + 1
                )
                all_xcorr[0, j] = corrval
                all_da[0, j] = shiftval / self.oversampling

            if plotmode:
                plt.show()

        # value with biggest cc value form table
        maximumcc = np.argmax(np.sum(all_xcorr, axis=1))
        dafinal = np.mean(all_da[maximumcc, :])
        for j in range(n_channels):
            index = self.group_index[j][group].nonzero()[1]
            if translateaxis == "x":
                self.locs[j].x[index] += dafinal
            elif translateaxis == "y":
                self.locs[j].y[index] += dafinal
            elif translateaxis == "z":
                self.locs[j].z[index] += dafinal * self.pixelsize

    def adjust_z(self):
        z_range_str = np.asarray((self.z_range.text()).split(","))
        z_range = []

        for element in z_range_str:
            try:
                z_range.append(float(element))
            except ValueError:
                pass

        z_min = z_range[0]
        z_max = z_range[1]

        self.z_min = np.max([z_min, self.z_min_load])
        self.z_max = np.min([z_max, self.z_max_load])

        print("Z min {}, Z max {}".format(self.z_min, self.z_max))

        self.updateLayout()

    def adjust_xy(self):
        x_range_str = np.asarray((self.x_range.text()).split(","))
        x_range = []

        for element in x_range_str:
            try:
                x_range.append(float(element))
            except ValueError:
                pass

        x_min = x_range[0]
        x_max = x_range[1]

        self.x_min = np.max([x_min, self.t_min])
        self.x_max = np.min([x_max, self.t_max])

        print("X min {}, X max {}".format(self.x_min, self.x_max))

        y_range_str = np.asarray((self.y_range.text()).split(","))
        y_range = []

        for element in y_range_str:
            try:
                y_range.append(float(element))
            except ValueError:
                pass

        y_min = y_range[0]
        y_max = y_range[1]

        self.y_min = np.max([y_min, self.t_min])
        self.y_max = np.min([y_max, self.t_max])

        print("Y min {}, Y max {}".format(self.y_min, self.y_max))

        self.updateLayout()

    def rotatexy_convolution_group(
        self, CF_image_avg, angles, group, rotaxis, proplane
    ):
        n_channels = len(self.locs)
        n_angles = len(angles)

        all_xcorr = np.zeros((n_angles, n_channels))
        all_da = np.zeros((n_angles, n_channels))
        all_db = np.zeros((n_angles, n_channels))

        for j in range(n_channels):
            if self.dataset_dialog.checks[j].isChecked():
                index = self.group_index[j][group].nonzero()[1]
                x_rot = self.locs[j].x[index]
                y_rot = self.locs[j].y[index]
                z_rot = self.locs[j].z[index]
                x_original = x_rot.copy()
                y_original = y_rot.copy()
                z_original = z_rot.copy()

                if self.translatebtn.isChecked():
                    angles = [0]
                    n_angles = 1

                for k in range(n_angles):
                    angle = angles[k]
                    # rotate locs
                    x_rot, y_rot, z_rot = rotate_axis(
                        rotaxis,
                        x_original,
                        y_original,
                        z_original,
                        angle,
                        self.pixelsize,
                    )
                    # render group image for plane
                    image = self.render_planes(
                        x_rot, y_rot, z_rot, proplane, self.pixelsize
                    )

                    # calculate cross-correlation
                    if 0:
                        fig = plt.figure()
                        ax1 = fig.add_subplot(1, 2, 1)
                        ax1.set_aspect("equal")
                        plt.imshow(image, interpolation="nearest", cmap=plt.cm.ocean)
                        plt.colorbar()
                        plt.show()
                        plt.waitforbuttonpress()

                    xcorr = np.sum(np.multiply(CF_image_avg[j], image))
                    all_xcorr[k, j] = xcorr

        # value with biggest cc value form table
        maximumcc = np.argmax(np.sum(all_xcorr, axis=1))
        rotfinal = angles[maximumcc]

        for j in range(n_channels):
            index = self.group_index[j][group].nonzero()[1]
            x_rot = self.locs[j].x[index]
            y_rot = self.locs[j].y[index]
            z_rot = self.locs[j].z[index]
            x_original = x_rot.copy()
            y_original = y_rot.copy()
            z_original = z_rot.copy()
            # rotate and shift image group locs
            x_rot, y_rot, z_rot = rotate_axis(
                rotaxis,
                x_original,
                y_original,
                z_original,
                rotfinal,
                self.pixelsize,
            )

            self.locs[j].x[index] = x_rot
            self.locs[j].y[index] = y_rot
            self.locs[j].z[index] = z_rot

    def rotatexy_convolution(self):
        # TODO: re-write ths with kwargs at some point
        rotaxis = []
        if self.x_axisbtn.isChecked():
            rotaxis = "x"
        elif self.y_axisbtn.isChecked():
            rotaxis = "y"
        elif self.z_axisbtn.isChecked():
            rotaxis = "z"

        n_groups = self.group_index[0].shape[0]
        a_step = np.arcsin(1 / (self.oversampling * self.r))
        if self.full_degbtn.isChecked():
            angles = np.arange(0, 2 * np.pi, a_step)
        elif self.part_degbtn.isChecked():
            degree = self.degEdit.value()
            angles = np.arange(
                -degree / 360 * 2 * np.pi, degree / 360 * 2 * np.pi, a_step
            )

        renderings = [
            render.render_hist3d(
                _,
                self.oversampling,
                self.t_min,
                self.t_min,
                self.t_max,
                self.t_max,
                self.z_min,
                self.z_max,
                self.pixelsize,
            )
            for _ in self.locs
        ]

        images = np.array([_[1] for _ in renderings])

        # DELIVER CORRECT PROJECTION FOR IMAGE
        proplane = []

        if self.xy_projbtn.isChecked():
            proplane = "xy"
            image = [np.sum(_, axis=2) for _ in images]
        elif self.yz_projbtn.isChecked():
            proplane = "yz"
            image = [np.sum(_, axis=1) for _ in images]
            image = [_.transpose() for _ in image]
        elif self.xz_projbtn.isChecked():
            proplane = "xz"
            image = [(np.sum(_, axis=0)) for _ in images]
            image = [_.transpose() for _ in image]

        # Change CFiamge for symmetry
        if self.radio_sym.isChecked():
            print("Using symmetry.")
            fig = plt.figure(figsize=(5, 5))
            ax1 = fig.add_subplot(1, 2, 1)
            symmetry = self.symEdit.value()
            ax1.set_aspect("equal")
            imageold = image[0].copy()
            plt.imshow(imageold, interpolation="nearest", cmap=plt.cm.ocean)

            # rotate image
            for i in range(symmetry - 1):
                image[0] += scipy.ndimage.interpolation.rotate(
                    imageold,
                    ((i + 1) * 360 / symmetry),
                    axes=(1, 0),
                    reshape=False,
                )

            ax2 = fig.add_subplot(1, 2, 2)
            ax2.set_aspect("equal")
            plt.imshow(image[0], interpolation="nearest", cmap=plt.cm.ocean)
            fig.canvas.draw()
            size = fig.canvas.size()
            width, height = size.width(), size.height()
            im = QtGui.QImage(
                fig.canvas.buffer_rgba(),
                width,
                height,
                QtGui.QImage.Format_ARGB32,
            )
            self.viewcp.setPixmap((QtGui.QPixmap(im)))
            self.viewcp.setAlignment(QtCore.Qt.AlignCenter)
            plt.close(fig)

        if self.radio_sym_custom.isChecked():
            print("Using custom symmetry.")
            symmetry_txt = np.asarray((self.symcustomEdit.text()).split(","))
            print(symmetry_txt)
            fig = plt.figure(figsize=(5, 5))
            ax1 = fig.add_subplot(1, 2, 1)
            symmetry = self.symEdit.value()
            ax1.set_aspect("equal")
            imageold = image[0].copy()
            plt.imshow(imageold, interpolation="nearest", cmap=plt.cm.ocean)

            # rotate image
            for degree in symmetry_txt:
                image[0] += scipy.ndimage.interpolation.rotate(
                    imageold, float(degree), axes=(1, 0), reshape=False
                )

            ax2 = fig.add_subplot(1, 2, 2)
            ax2.set_aspect("equal")
            plt.imshow(image[0], interpolation="nearest", cmap=plt.cm.ocean)
            fig.canvas.draw()
            size = fig.canvas.size()
            width, height = size.width(), size.height()
            im = QtGui.QImage(
                fig.canvas.buffer_rgba(),
                width,
                height,
                QtGui.QImage.Format_ARGB32,
            )
            self.viewcp.setPixmap((QtGui.QPixmap(im)))
            self.viewcp.setAlignment(QtCore.Qt.AlignCenter)
            plt.close(fig)

        if self.modelchk.isChecked():
            self.generate_template()
            image[0] = self.template_img

        CF_image_avg = image

        print("Convolving..")
        for i in range(n_groups):
            self.status_bar.showMessage("Group {} / {}.".format(i, n_groups))
            self.rotatexy_convolution_group(CF_image_avg, angles, i, rotaxis, proplane)
        self.updateLayout()
        self.status_bar.showMessage("Done!")

    def rotate_groups(self):
        # Read out values from radiobuttons
        # TODO: maybe re-write this with kwargs
        rotaxis = []
        if self.x_axisbtn.isChecked():
            rotaxis = "x"
        elif self.y_axisbtn.isChecked():
            rotaxis = "y"
        elif self.z_axisbtn.isChecked():
            rotaxis = "z"

        n_groups = self.group_index[0].shape[0]
        a_step = np.arcsin(1 / (self.oversampling * self.r))

        if self.full_degbtn.isChecked():
            angles = np.arange(0, 2 * np.pi, a_step)
        elif self.part_degbtn.isChecked():
            degree = self.degEdit.value()
            angles = np.arange(
                -degree / 360 * 2 * np.pi, degree / 360 * 2 * np.pi, a_step
            )

        renderings = [
            render.render_hist3d(
                _,
                self.oversampling,
                self.t_min,
                self.t_min,
                self.t_max,
                self.t_max,
                self.z_min,
                self.z_max,
                self.pixelsize,
            )
            for _ in self.locs
        ]
        images = np.array([_[1] for _ in renderings])

        # DELIVER CORRECT PROJECTION FOR IMAGE
        proplane = []
        if self.xy_projbtn.isChecked():

            proplane = "xy"
            image = [np.sum(_, axis=2) for _ in images]
        elif self.yz_projbtn.isChecked():

            proplane = "yz"

            image = [np.sum(_, axis=1) for _ in images]
            image = [_.transpose() for _ in image]
        elif self.xz_projbtn.isChecked():

            proplane = "xz"
            image = [(np.sum(_, axis=0)) for _ in images]
            image = [_.transpose() for _ in image]

        if self.radio_sym.isChecked():
            print("Radio sym")
            fig = plt.figure(figsize=(5, 5))
            ax1 = fig.add_subplot(1, 2, 1)
            symmetry = self.symEdit.value()
            ax1.set_aspect("equal")
            imageold = image[0].copy()
            plt.imshow(imageold, interpolation="nearest", cmap=plt.cm.ocean)

            # rotate image
            for i in range(symmetry - 1):
                image[0] += scipy.ndimage.interpolation.rotate(
                    imageold,
                    ((i + 1) * 360 / symmetry),
                    axes=(1, 0),
                    reshape=False,
                )

            ax2 = fig.add_subplot(1, 2, 2)
            ax2.set_aspect("equal")
            plt.imshow(image[0], interpolation="nearest", cmap=plt.cm.ocean)
            fig.canvas.draw()
            size = fig.canvas.size()
            width, height = size.width(), size.height()
            im = QtGui.QImage(
                fig.canvas.buffer_rgba(),
                width,
                height,
                QtGui.QImage.Format_ARGB32,
            )
            self.viewcp.setPixmap((QtGui.QPixmap(im)))
            self.viewcp.setAlignment(QtCore.Qt.AlignCenter)
            plt.close(fig)

        # TODO: Sort these functions out,
        # combine with radio_sym / also for convolving.
        if self.radio_sym_custom.isChecked():
            print("Using custom symmetry.")
            symmetry_txt = np.asarray((self.symcustomEdit.text()).split(","))
            fig = plt.figure(figsize=(5, 5))
            ax1 = fig.add_subplot(1, 2, 1)
            symmetry = self.symEdit.value()
            ax1.set_aspect("equal")
            imageold = image[0].copy()
            plt.imshow(imageold, interpolation="nearest", cmap=plt.cm.ocean)

            # rotate image
            for degree in symmetry_txt:
                image[0] += scipy.ndimage.interpolation.rotate(
                    imageold, float(degree), axes=(1, 0), reshape=False
                )

            ax2 = fig.add_subplot(1, 2, 2)
            ax2.set_aspect("equal")
            plt.imshow(image[0], interpolation="nearest", cmap=plt.cm.ocean)
            fig.canvas.draw()
            size = fig.canvas.size()
            width, height = size.width(), size.height()
            im = QtGui.QImage(
                fig.canvas.buffer_rgba(),
                width,
                height,
                QtGui.QImage.Format_ARGB32,
            )
            self.viewcp.setPixmap((QtGui.QPixmap(im)))
            self.viewcp.setAlignment(QtCore.Qt.AlignCenter)
            plt.close(fig)

        if self.modelchk.isChecked():
            self.generate_template()
            image[0] = self.template_img

        CF_image_avg = [np.conj(np.fft.fft2(_)) for _ in image]
        # n_pixel, _ = image_avg.shape
        # image_half = n_pixel / 2

        print("Rotating..")
        for i in range(n_groups):
            self.status_bar.showMessage("Group {} / {}.".format(i, n_groups))
            self.align_group(CF_image_avg, angles, i, rotaxis, proplane)
        self.updateLayout()
        self.status_bar.showMessage("Done!")

    def getUIstate(self):
        rotaxis = []
        if self.x_axisbtn.isChecked():
            rotaxis = "x"
        elif self.y_axisbtn.isChecked():
            rotaxis = "y"
        elif self.z_axisbtn.isChecked():
            rotaxis = "z"

        proplane = []
        if self.xy_projbtn.isChecked():
            proplane = "xy"
        elif self.yz_projbtn.isChecked():
            proplane = "yz"
        elif self.xz_projbtn.isChecked():
            proplane = "xz"

        return rotaxis, proplane

    def projectPlanes(self, images, proplane):
        if proplane == "xy":
            image = [np.sum(_, axis=2) for _ in images]
        elif proplane == "yz":
            image = [np.sum(_, axis=1) for _ in images]
            image = [_.transpose() for _ in image]
        elif proplane == "xz":
            image = [(np.sum(_, axis=0)) for _ in images]
            image = [_.transpose() for _ in image]

        return image

    def generate_template(self):
        model_x_str = np.asarray((self.model_x.text()).split(","))
        model_y_str = np.asarray((self.model_y.text()).split(","))
        model_z_str = np.asarray((self.model_z.text()).split(","))

        model_x = []
        model_y = []
        model_z = []

        for element in model_x_str:
            try:
                model_x.append(float(element))
            except ValueError:
                pass

        for element in model_y_str:
            try:
                model_y.append(float(element))
            except ValueError:
                pass

        for element in model_z_str:
            try:
                model_z.append(float(element))
            except ValueError:
                pass

        pixelsize = self.pixelsizeEdit.value()
        blur = self.modelblurEdit.value()

        # Center of mass
        model_x = np.array(model_x) / pixelsize
        model_y = np.array(model_y) / pixelsize
        model_z = np.array(model_z)

        model_x = model_x - np.mean(model_x)
        model_y = model_y - np.mean(model_y)
        model_z = model_z - np.mean(model_z)

        rotaxis, proplane = self.getUIstate()

        template_img = self.render_planes(
            model_x, model_y, model_z, proplane, pixelsize
        )

        self.template_img = scipy.ndimage.filters.gaussian_filter(template_img, blur)

    def model_preview(self):

        self.generate_template()
        # Generate a  template image

        fig = plt.figure()
        plt.title("Preview of Template")
        plt.imshow(self.template_img, interpolation="nearest", cmap=plt.cm.hot)
        plt.show()

    def calculate_score(self):
        # Dummy button -> Functionality of rotatebtn for now
        # TODO: maybe re-write this with kwargs
        self.scores = []
        rotaxis, proplane = self.getUIstate()
        n_groups = self.group_index[0].shape[0]

        renderings = [
            render.render_hist3d(
                _,
                self.oversampling,
                self.t_min,
                self.t_min,
                self.t_max,
                self.t_max,
                self.z_min,
                self.z_max,
                self.pixelsize,
            )
            for _ in self.locs
        ]
        n_locs = sum([_[0] for _ in renderings])
        # Make an average and not a sum image here..
        images = np.array([_[1] / n_groups for _ in renderings])

        # DELIVER CORRECT PROJECTION FOR IMAGE
        image = self.projectPlanes(images, proplane)

        n_channels = len(image)

        print("Calculating score..")
        for i in range(n_groups):
            channel_score = []
            for j in range(n_channels):
                if self.dataset_dialog.checks[j].isChecked():
                    index = self.group_index[j][i].nonzero()[1]
                    x_rot = self.locs[j].x[index]
                    y_rot = self.locs[j].y[index]
                    z_rot = self.locs[j].z[index]

                    groupimage = self.render_planes(
                        x_rot, y_rot, z_rot, proplane, self.pixelsize
                    )
                    score = np.sum(np.sqrt(groupimage * image[j])) / np.sum(
                        np.sqrt(groupimage * groupimage)
                    )
                    channel_score.append(score)
                self.scores.append(channel_score)

            self.status_bar.showMessage("Group {} / {}.".format(i, n_groups))
        self.status_bar.showMessage(
            "Done. Average score: {}".format(np.mean(self.scores))
        )

        plt.hist(np.array(self.scores), 40)
        plt.title("Histogram of Scores, Mean: {:.2f}".format(np.mean(self.scores)))
        plt.xlabel("Score")
        plt.ylabel("Counts")
        plt.show()

    def mean_angle(self, deg):
        return phase(sum(rect(1, d) for d in deg) / len(deg))

    def render_planes(self, xdata, ydata, zdata, proplane, pixelsize):
        # assign correct renderings for all planes
        a_render = []
        b_render = []

        if proplane == "xy":
            a_render = xdata
            b_render = ydata
            aval_min = self.t_min
            aval_max = self.t_max
            bval_min = self.t_min
            bval_max = self.t_max
        elif proplane == "yz":
            a_render = ydata
            b_render = np.divide(zdata, pixelsize)
            aval_min = self.t_min
            aval_max = self.t_max
            bval_min = np.divide(self.z_min, pixelsize)
            bval_max = np.divide(self.z_max, pixelsize)
        elif proplane == "xz":
            b_render = np.divide(zdata, pixelsize)
            a_render = xdata
            bval_min = np.divide(self.z_min, pixelsize)
            bval_max = np.divide(self.z_max, pixelsize)
            aval_min = self.t_min
            aval_max = self.t_max

        N, plane = render_histxyz(
            a_render,
            b_render,
            self.oversampling,
            aval_min,
            aval_max,
            bval_min,
            bval_max,
        )
        return plane

    def align_all(self, alignaxis):
        a_step = np.arcsin(1 / (self.oversampling * self.r))
        angles = np.arange(0, 2 * np.pi, a_step)
        n_channels = len(self.locs)
        n_angles = len(angles)
        all_corr = np.zeros((n_angles, n_channels))

        for j in range(n_channels):
            if self.dataset_dialog.checks[j].isChecked():
                alignimage = []
                x_rot = self.locs[j].x
                y_rot = self.locs[j].y
                z_rot = self.locs[j].z

                x_original = x_rot.copy()
                y_original = y_rot.copy()
                z_original = z_rot.copy()

                alignimage = []

                for k in range(n_angles):
                    angle = angles[k]
                    if alignaxis == "zz":
                        proplane = "yz"
                        rotaxis = "x"
                    elif alignaxis == "zy":
                        proplane = "yz"
                        rotaxis = "x"
                    elif alignaxis == "y":
                        proplane = "xy"
                        rotaxis = "z"
                    elif alignaxis == "x":
                        proplane = "xy"
                        rotaxis = "z"

                    x_rot, y_rot, z_rot = rotate_axis(
                        rotaxis,
                        x_original,
                        y_original,
                        z_original,
                        angle,
                        self.pixelsize,
                    )
                    # render group image for plane
                    image = self.render_planes(
                        x_rot, y_rot, z_rot, proplane, self.pixelsize
                    )  # RENDR PLANES WAS BUGGY AT SOME POINT

                    if alignimage == []:
                        alignimage = np.zeros(image.shape)
                        # CREATE ALIGNIMAGE
                        if alignaxis == "zz":
                            alignimage[np.int64(alignimage.shape[0] / 2), :] += 2
                            alignimage[np.int64(alignimage.shape[0] / 2) + 1, :] += 1
                            alignimage[np.int64(alignimage.shape[0] / 2) - 1, :] += 1
                        elif alignaxis == "zy":
                            alignimage[:, np.int64(alignimage.shape[0] / 2)] += 2
                            alignimage[:, np.int64(alignimage.shape[0] / 2) + 1] += 1
                            alignimage[:, np.int64(alignimage.shape[0] / 2) - 1] += 1
                        elif alignaxis == "y":
                            alignimage[:, np.int64(alignimage.shape[1] / 2)] += 2
                            alignimage[:, np.int64(alignimage.shape[1] / 2) - 1] += 1
                            alignimage[:, np.int64(alignimage.shape[1] / 2) + 1] += 1
                        elif alignaxis == "x":
                            alignimage[np.int64(alignimage.shape[0] / 2), :] += 2
                            alignimage[np.int64(alignimage.shape[0] / 2) + 1, :] += 1
                            alignimage[np.int64(alignimage.shape[0] / 2) - 1, :] += 1

                    all_corr[k, j] = np.sum(np.multiply(alignimage, image))

                    if 0:
                        fig = plt.figure()
                        ax1 = fig.add_subplot(1, 2, 1)
                        ax1.set_aspect("equal")
                        plt.imshow(image, interpolation="nearest", cmap=plt.cm.ocean)
                        ax2 = fig.add_subplot(1, 2, 2)
                        ax2.set_aspect("equal")
                        plt.imshow(
                            alignimage,
                            interpolation="nearest",
                            cmap=plt.cm.ocean,
                        )
                        plt.colorbar()
                        plt.show()

        # value with biggest cc value form table
        maximumcc = np.argmax(np.sum(all_corr, axis=1))
        rotfinal = angles[maximumcc]

        for j in range(n_channels):
            x_rot = self.locs[j].x
            y_rot = self.locs[j].y
            z_rot = self.locs[j].z
            x_original = x_rot.copy()
            y_original = y_rot.copy()
            z_original = z_rot.copy()
            # rotate and shift image group locs

            x_rot, y_rot, z_rot = rotate_axis(
                rotaxis,
                x_original,
                y_original,
                z_original,
                rotfinal,
                self.pixelsize,
            )
            self.locs[j].x = x_rot
            self.locs[j].y = y_rot
            self.locs[j].z = z_rot

        self.updateLayout()
        self.status_bar.showMessage("Align on Axis {} complete.".format(alignaxis))

    def align_group(self, CF_image_avg, angles, group, rotaxis, proplane):
        n_channels = len(self.locs)

        n_angles = len(angles)

        all_xcorr = np.zeros((n_angles, n_channels))
        all_da = np.zeros((n_angles, n_channels))
        all_db = np.zeros((n_angles, n_channels))

        flips = 1
        if self.flipbtn.isChecked():
            print("Considering flipped structures...")
            flips = 2

        for f in range(flips):
            for j in range(n_channels):
                if self.dataset_dialog.checks[j].isChecked():
                    index = self.group_index[j][group].nonzero()[1]
                    x_rot = self.locs[j].x[index]
                    y_rot = self.locs[j].y[index]
                    z_rot = self.locs[j].z[index]
                    x_original = x_rot.copy()
                    y_original = y_rot.copy()
                    z_original = z_rot.copy()

                    if f == 1:  # Flipped round
                        if proplane == "xy":
                            x_original = -x_original
                        elif proplane == "yz":
                            y_original = -y_original
                        elif proplane == "xz":
                            z_original = -z_original

                    if self.translatebtn.isChecked():
                        angles = [0]
                        n_angles = 1

                    for k in range(n_angles):
                        angle = angles[k]

                        x_rot, y_rot, z_rot = rotate_axis(
                            rotaxis,
                            x_original,
                            y_original,
                            z_original,
                            angle,
                            self.pixelsize,
                        )
                        # render group image for plane
                        image = self.render_planes(
                            x_rot, y_rot, z_rot, proplane, self.pixelsize
                        )  # RENDR PLANES WAS BUGGY AT SOME POINT

                        # calculate cross-correlation
                        if 0:
                            fig = plt.figure()
                            ax1 = fig.add_subplot(1, 2, 1)
                            ax1.set_aspect("equal")
                            plt.imshow(
                                image,
                                interpolation="nearest",
                                cmap=plt.cm.ocean,
                            )
                            plt.colorbar()
                            plt.show()
                            plt.waitforbuttonpress()

                        xcorr = compute_xcorr(CF_image_avg[j], image)

                        n_pixelb, n_pixela = image.shape
                        image_halfa = n_pixela / 2  # TODO: CHECK THOSE VALUES
                        image_halfb = n_pixelb / 2

                        # find the brightest pixel
                        b_max, a_max = np.unravel_index(xcorr.argmax(), xcorr.shape)
                        # store the transformation if the correlation
                        # is larger than before
                        all_xcorr[k, j] = xcorr[b_max, a_max]
                        all_db[k, j] = np.ceil(b_max - image_halfb) / self.oversampling
                        all_da[k, j] = np.ceil(a_max - image_halfa) / self.oversampling

            flipstate = False
            if f == 0:
                # value with biggest cc value form table
                maximumcc = np.argmax(np.sum(all_xcorr, axis=1))
                maximumcc_val = np.max(np.sum(all_xcorr, axis=1))
                rotfinal = angles[maximumcc]

                dafinal = np.mean(all_da[maximumcc, :])
                dbfinal = np.mean(all_db[maximumcc, :])

            else:
                maximumcc_val_f = np.max(np.sum(all_xcorr, axis=1))

                if maximumcc_val < maximumcc_val_f:
                    flipstate = True
                    maximumcc = np.argmax(np.sum(all_xcorr, axis=1))
                    rotfinal = angles[maximumcc]

                    dafinal = np.mean(all_da[maximumcc, :])
                    dbfinal = np.mean(all_db[maximumcc, :])

        for j in range(n_channels):
            index = self.group_index[j][group].nonzero()[1]

            x_rot = self.locs[j].x[index]
            y_rot = self.locs[j].y[index]
            z_rot = self.locs[j].z[index]
            x_original = x_rot.copy()
            y_original = y_rot.copy()
            z_original = z_rot.copy()
            # rotate and shift image group locs
            x_rot, y_rot, z_rot = rotate_axis(
                rotaxis,
                x_original,
                y_original,
                z_original,
                rotfinal,
                self.pixelsize,
            )

            if flipstate:
                if proplane == "xy":
                    self.locs[j].x[index] = -x_rot
                    self.locs[j].y[index] = y_rot
                    self.locs[j].z[index] = z_rot
                elif proplane == "yz":
                    self.locs[j].x[index] = x_rot
                    self.locs[j].y[index] = -y_rot
                    self.locs[j].z[index] = z_rot
                elif proplane == "xz":
                    self.locs[j].x[index] = x_rot
                    self.locs[j].y[index] = y_rot
                    self.locs[j].z[index] = -z_rot

            else:
                self.locs[j].x[index] = x_rot
                self.locs[j].y[index] = y_rot
                self.locs[j].z[index] = z_rot

            # Shift image group locs
            if self.translatebtn.isChecked():
                dbfinal = 0
            if proplane == "xy":
                self.locs[j].x[index] -= dafinal
                self.locs[j].y[index] -= dbfinal
            elif proplane == "yz":
                self.locs[j].y[index] -= dafinal
                self.locs[j].z[index] -= dbfinal * self.pixelsize
            elif proplane == "xz":
                self.locs[j].z[index] -= dafinal
                self.locs[j].x[index] -= dbfinal * self.pixelsize

    def fit_in_view(self, autoscale=False):
        movie_height, movie_width = self.movie_size()
        viewport = [(0, 0), (movie_height, movie_width)]
        self.update_scene(viewport=viewport, autoscale=autoscale)

    def movie_size(self):
        movie_height = self.max_movie_height()
        movie_width = self.max_movie_width()
        return (movie_height, movie_width)

    def max_movie_height(self):
        """Returns maximum height of all loaded images."""
        return max(info[0]["Height"] for info in self.infos)

    def max_movie_width(self):
        return max([info[0]["Width"] for info in self.infos])

    def update_scene(
        self, viewport=None, autoscale=False, use_cache=False, picks_only=False
    ):
        n_channels = len(self.locs)
        if n_channels:
            viewport = viewport or self.viewport
            self.draw_scene(
                viewport,
                autoscale=autoscale,
                use_cache=use_cache,
                picks_only=picks_only,
            )
            # self.update_cursor()

    def draw_scene(self, viewport, autoscale=False, use_cache=False, picks_only=False):
        self.viewport = self.adjust_viewport_to_view(viewport)
        qimage = self.render_scene(autoscale=autoscale, use_cache=use_cache)
        self.qimage = qimage.scaled(
            self.viewxy.width(),
            self.viewxy.height(),
            QtCore.Qt.KeepAspectRatioByExpanding,
        )

    def adjust_viewport_to_view(self, viewport):
        """
        Adds space to a desired viewport so that
        it matches the window aspect ratio.
        """
        viewport_height = viewport[1][0] - viewport[0][0]
        viewport_width = viewport[1][1] - viewport[0][1]
        view_height = self.height()
        view_width = self.width()
        viewport_aspect = viewport_width / viewport_height
        view_aspect = view_width / view_height
        if view_aspect >= viewport_aspect:
            y_min = viewport[0][0]
            y_max = viewport[1][0]
            x_range = viewport_height * view_aspect
            x_margin = (x_range - viewport_width) / 2
            x_min = viewport[0][1] - x_margin
            x_max = viewport[1][1] + x_margin
        else:
            x_min = viewport[0][1]
            x_max = viewport[1][1]
            y_range = viewport_width / view_aspect
            y_margin = (y_range - viewport_height) / 2
            y_min = viewport[0][0] - y_margin
            y_max = viewport[1][0] + y_margin
        return [(y_min, x_min), (y_max, x_max)]

    def render_scene(self, autoscale=False, use_cache=False, cache=True, viewport=None):
        kwargs = self.get_render_kwargs(viewport=viewport)
        n_channels = len(self.locs)
        if n_channels == 1:
            self.render_single_channel(
                kwargs, autoscale=autoscale, use_cache=use_cache, cache=cache
            )
        else:
            self.render_multi_channel(
                kwargs, autoscale=autoscale, use_cache=use_cache, cache=cache
            )
        self._bgra[:, :, 3].fill(255)
        Y, X = self._bgra.shape[:2]
        qimage = QtGui.QImage(self._bgra.data, X, Y, QtGui.QImage.Format_RGB32)
        return qimage

    def get_render_kwargs(self, viewport=None):  # Dummy for now: TODO: Implement
        viewport = [(0, 0), (32, 32)]
        return {
            "oversampling": 5,
            "viewport": viewport,
            "blur_method": None,
            "min_blur_width": float(0),
        }

    def render_multi_channel(
        self, kwargs, autoscale=False, locs=None, use_cache=False, cache=True
    ):
        if locs is None:
            locs = self.locs
        n_channels = len(locs)
        hues = np.arange(0, 1, 1 / n_channels)
        colors = [colorsys.hsv_to_rgb(_, 1, 1) for _ in hues]
        if use_cache:
            n_locs = self.n_locs
            image = self.image
        else:
            renderings = [render.render(_, **kwargs) for _ in locs]
            n_locs = sum([_[0] for _ in renderings])
            image = np.array([_[1] for _ in renderings])
        if cache:
            self.n_locs = n_locs
            self.image = image
        image = self.scale_contrast(image)
        Y, X = image.shape[1:]
        bgra = np.zeros((Y, X, 4), dtype=np.float32)
        for color, image in zip(colors, image):
            bgra[:, :, 0] += color[2] * image
            bgra[:, :, 1] += color[1] * image
            bgra[:, :, 2] += color[0] * image
        bgra = np.minimum(bgra, 1)
        self._bgra = self.to_8bit(bgra)
        return self._bgra

    def render_single_channel(
        self, kwargs, autoscale=False, use_cache=False, cache=True
    ):
        locs = self.locs[0]
        if hasattr(locs, "group"):
            locs = [locs[self.group_color == _] for _ in range(N_GROUP_COLORS)]
            return self.render_multi_channel(
                kwargs, autoscale=autoscale, locs=locs, use_cache=use_cache
            )
        if use_cache:
            n_locs = self.n_locs
            image = self.image
        else:
            n_locs, image = render.render(locs, **kwargs)
        if cache:
            self.n_locs = n_locs
            self.image = image
        image = self.scale_contrast(image, autoscale=autoscale)
        image = self.to_8bit(image)
        Y, X = image.shape
        # cmap = self.window.display_settings_dialog.colormap.currentText()
        # TODO: selection of colormap?
        cmap = "hot"
        cmap = np.uint8(np.round(255 * plt.get_cmap(cmap)(np.arange(256))))
        self._bgra = np.zeros((Y, X, 4), dtype=np.uint8, order="C")
        self._bgra[..., 0] = cmap[:, 2][image]
        self._bgra[..., 1] = cmap[:, 1][image]
        self._bgra[..., 2] = cmap[:, 0][image]
        return self._bgra

    def to_8bit(self, image):
        return np.round(255 * image).astype("uint8")

    def scale_contrast(self, image, autoscale=False):

        if image.ndim == 2:
            max_ = image.max()
        else:
            max_ = min([_.max() for _ in image])
        upper = self.contrastEdit.value() * max_

        lower = 0

        if upper > 0:
            image = (image - lower) / (upper - lower)
            image[~np.isfinite(image)] = 0
            image = np.minimum(image, 1.0)
            image = np.maximum(image, 0.0)
        return image


def main():

    app = QtWidgets.QApplication(sys.argv)
    window = Window()

    from . import plugins

    def iter_namespace(pkg):
        return pkgutil.iter_modules(pkg.__path__, pkg.__name__ + ".")

    plugins = [
        importlib.import_module(name)
        for finder, name, ispkg
        in iter_namespace(plugins)
    ]

    for plugin in plugins:
        p = plugin.Plugin(window)
        if p.name == "average3":
            p.execute()

    window.show()

    def excepthook(type, value, tback):
        lib.cancel_dialogs()
        message = "".join(traceback.format_exception(type, value, tback))
        errorbox = QtWidgets.QMessageBox.critical(window, "An error occured", message)
        errorbox.exec_()
        sys.__excepthook__(type, value, tback)

    sys.excepthook = excepthook

    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



"""
    picasso.imageprocess
    ~~~~~~~~~~~~~~~~~~~~

    Image processing functions

    :author: Joerg Schnitzbauer, 2016
    :copyright: Copyright (c) 2016 Jungmann Lab, MPI of Biochemistry
"""

from typing import Callable

import matplotlib.pyplot as _plt
import numpy as _np
from numpy import fft as _fft
import lmfit as _lmfit
from tqdm import tqdm as _tqdm
from . import lib as _lib
from . import render as _render
from . import localize as _localize
from . import postprocess as _postprocess

_plt.style.use("ggplot")


def xcorr(imageA: _np.ndarray, imageB: _np.ndarray) -> _np.ndarray:
    """Computes the cross-correlation of two images using FFT.
    
    Parameters
    ----------
    imageA, imageB : np.ndarray
        Input images to be cross-correlated. They should have the same 
        shape.
    
    Returns
    -------
    res : np.ndarray
        The cross-correlation result, which is the inverse Fourier
        transform of the product of the Fourier transforms of the two
        images.
    """

    FimageA = _fft.fft2(imageA)
    CFimageB = _np.conj(_fft.fft2(imageB))
    res = _fft.fftshift(_np.real(_fft.ifft2((FimageA * CFimageB)))) / _np.sqrt(
        imageA.size
    )
    return res


def get_image_shift(
    imageA: _np.ndarray, 
    imageB: _np.ndarray, 
    box: int, 
    roi: int | None = None, 
    display: bool = False,
) -> tuple[float, float]:
    """Computes the shift from imageA to imageB.
    
    Parameters
    ----------
    imageA, imageB : np.ndarray
        Input images to be cross-correlated. They should have the same
        shape.
    box : int
        Size of the box used for fitting the cross-correlation peak.
    roi : int, optional
        Region of interest size to cut out the center of the 
        cross-correlation image. If None, the entire cross-correlation
        image is used.
    display : bool, optional
        If True, displays the images and the cross-correlation result.
    
    Returns
    -------
    yc, xc : float
        The y and x coordinates of the shift from imageA to imageB.
        The coordinates are adjusted to be relative to the center of
        the images.
    """

    if (_np.sum(imageA) == 0) or (_np.sum(imageB) == 0):
        return 0, 0
    # Compute image correlation
    XCorr = xcorr(imageA, imageB)
    # Cut out center roi
    Y, X = imageA.shape
    if roi is not None:
        Y_ = int((Y - roi) / 2)
        X_ = int((X - roi) / 2)
        if Y_ > 0:
            XCorr = XCorr[Y_:-Y_, :]
        else:
            Y_ = 0
        if X_ > 0:
            XCorr = XCorr[:, X_:-X_]
        else:
            X_ = 0
    else:
        Y_ = X_ = 0
    # A quarter of the fit ROI
    fit_X = int(box / 2)
    # A coordinate grid for the fitting ROI
    y, x = _np.mgrid[-fit_X : fit_X + 1, -fit_X : fit_X + 1]
    # Find the brightest pixel and cut out the fit ROI
    y_max_, x_max_ = _np.unravel_index(XCorr.argmax(), XCorr.shape)
    FitROI = XCorr[
        y_max_ - fit_X : y_max_ + fit_X + 1,
        x_max_ - fit_X : x_max_ + fit_X + 1,
    ]

    dimensions = FitROI.shape

    if 0 in dimensions or dimensions[0] != dimensions[1]:
        xc, yc = 0, 0
    else:
        # The fit model
        def flat_2d_gaussian(a, xc, yc, s, b):
            A = a * _np.exp(-0.5 * ((x - xc) ** 2 + (y - yc) ** 2) / s**2) + b
            return A.flatten()

        gaussian2d = _lmfit.Model(
            flat_2d_gaussian, name="2D Gaussian", independent_vars=[]
        )

        # Set up initial parameters and fit
        params = _lmfit.Parameters()
        params.add("a", value=FitROI.max(), vary=True, min=0)
        params.add("xc", value=0, vary=True)
        params.add("yc", value=0, vary=True)
        params.add("s", value=1, vary=True, min=0)
        params.add("b", value=FitROI.min(), vary=True, min=0)
        results = gaussian2d.fit(FitROI.flatten(), params)

        # Get maximum coordinates and add offsets
        xc = results.best_values["xc"]
        yc = results.best_values["yc"]
        xc += X_ + x_max_
        yc += Y_ + y_max_

        if display:
            _plt.figure(figsize=(17, 10))
            _plt.subplot(1, 3, 1)
            _plt.imshow(imageA, interpolation="none")
            _plt.subplot(1, 3, 2)
            _plt.imshow(imageB, interpolation="none")
            _plt.subplot(1, 3, 3)
            _plt.imshow(XCorr, interpolation="none")
            _plt.plot(xc, yc, "x")
            _plt.show()

        xc -= _np.floor(X / 2)
        yc -= _np.floor(Y / 2)

    return -yc, -xc


def rcc(
    segments: list[_np.ndarray], 
    max_shift: float | None = None, 
    callback: Callable[[int], None] | None = None,
) -> _np.ndarray:
    """Computes RCC, see Wang, Schnitzbauer, et al. Optics Express, 
    2014. Returns the shifts in x and y directions for each pair of
    segments. 
    
    Parameters
    ----------
    segments : list of np.ndarray
        List of image segments to be correlated. Each segment should be
        a 2D numpy array representing an image.
    max_shift : float, optional
        Maximum allowed shift in pixels. If None, the default value is
        set to 5 pixels.
    callback : Callable[[int], None], optional
        A callback function that takes an integer argument and is called
        after processing each pair of segments. This can be used to
        update a progress bar or perform other actions during the
        computation.
    
    Returns
    -------
    shifts : np.ndarray
        A 2D numpy array containing the shifts in x and y directions for
        each pair of segments. The shape of the array is 
        (n_segments, n_segments), where n_segments is the number of 
        segments provided.
    """

    n_segments = len(segments)
    shifts_x = _np.zeros((n_segments, n_segments))
    shifts_y = _np.zeros((n_segments, n_segments))
    n_pairs = int(n_segments * (n_segments - 1) / 2)
    flag = 0
    if callback is None:
        with _tqdm(
            total=n_pairs, desc="Correlating image pairs", unit="pairs"
        ) as progress_bar:
            for i in range(n_segments - 1):
                for j in range(i + 1, n_segments):
                    progress_bar.update()
                    shifts_y[i, j], shifts_x[i, j] = get_image_shift(
                        segments[i], segments[j], 5, max_shift
                    )
                    flag += 1
    else:
        callback(0)
        for i in range(n_segments - 1):
            for j in range(i + 1, n_segments):
                shifts_y[i, j], shifts_x[i, j] = get_image_shift(
                    segments[i], segments[j], 5, max_shift
                )
                flag += 1
                callback(flag)
        
    return _lib.minimize_shifts(shifts_x, shifts_y)


def find_fiducials(
    locs: _np.recarray, 
    info: list[dict],
) -> tuple[list[tuple[int, int]], int]:
    """Finds the xy coordinates of regions with high density of 
    localizations, likely originating from fiducial markers. 

    Uses picasso.localize.identify_in_image with threshold set to 99th
    percentile of the image histogram. The image is rendered using 
    one-pixel-blur, see picasso.render.render.

    
    Parameters
    ----------
    locs : _np.recarray
        Localizations.
    info : list of dicts
        Localizations' metadata (from the corresponding .yaml file).
        
    Returns
    -------
    picks : list of (2,) tuples
        Coordinates of fiducial markers. Each list element corresponds 
        to (x, y) coordinates of one fiducial marker.
    box : int
        Size of the box used for the fiducial marker identification.
        Can be set as the pick diameter in pixels for undrifting.
    """

    image = _render.render(
        locs=locs,
        info=info,
        oversampling=1,
        viewport=None,
        blur_method="smooth",        
    )[1]
    # hist = _np.histogram(image.flatten(), bins=256)
    threshold = _np.percentile(image.flatten(), 99)
    # box size should be an odd number, corresponding to approximately
    # 900 nm 
    pixelsize = 130
    for inf in info:
        if val := inf.get("Pixelsize"):
            pixelsize = val
            break
    box = int(_np.round(900 / pixelsize))
    box = box + 1 if box % 2 == 0 else box

    # find the local maxima and translate to pick coordinates
    y, x, _ = _localize.identify_in_image(image, threshold, box=box)
    picks = [(xi, yi) for xi, yi in zip(x, y)]

    # select the picks with appropriate number of localizations
    n_frames = 0
    for inf in info:
        if val := inf.get("Frames"):
            n_frames = val
            break
    min_n = 0.8 * n_frames
    picked_locs = _postprocess.picked_locs(
        locs, info, picks, "Circle", pick_size=box/2, add_group=False,
    )
    picks = [
        pick for i, pick in enumerate(picks) if len(picked_locs[i]) > min_n
    ]
    return picks, box


"""
    picasso.localize
    ~~~~~~~~~~~~~~~~

    Identify and localize fluorescent single molecules in a frame sequence

    :authors: Joerg Schnitzbauer, Maximilian Thomas Strauss, 2016-2018
    :copyright: Copyright (c) 2016-2018 Jungmann Lab, MPI of Biochemistry
"""

from __future__ import annotations
from typing import Literal
from typing import Callable

import numpy as _np
import dask.array as _da
import numba as _numba
import multiprocessing as _multiprocessing
import ctypes as _ctypes
from concurrent.futures import ThreadPoolExecutor as _ThreadPoolExecutor
import threading as _threading
from itertools import chain as _chain
import matplotlib.pyplot as _plt
from . import gaussmle as _gaussmle
from . import io as _io
from . import postprocess as _postprocess
from . import __main__ as main
import os
from datetime import datetime
from sqlalchemy import create_engine
import pandas as pd

MAX_LOCS = int(1e6)

_C_FLOAT_POINTER = _ctypes.POINTER(_ctypes.c_float)
LOCS_DTYPE = [
    ("frame", "u4"),
    ("x", "f4"),
    ("y", "f4"),
    ("photons", "f4"),
    ("sx", "f4"),
    ("sy", "f4"),
    ("bg", "f4"),
    ("lpx", "f4"),
    ("lpy", "f4"),
    ("net_gradient", "f4"),
    ("likelihood", "f4"),
    ("iterations", "i4"),
]

MEAN_COLS = [
    "frame",
    "x",
    "y",
    "photons",
    "sx",
    "sy",
    "bg",
    "lpx",
    "lpy",
    "ellipticity",
    "net_gradient",
    "z",
    "d_zcalib",
]
SET_COLS = ["Frames", "Height", "Width", "Box Size", "Min. Net Gradient", "Pixelsize"]

_plt.style.use("ggplot")


@_numba.jit(nopython=True, nogil=True, cache=False)
def local_maxima(
    frame: _np.ndarray, 
    box: int
) -> tuple[_np.ndarray, _np.ndarray]:
    """Finds pixels with maximum value within a region of interest.
    
    Parameters
    ----------
    frame : np.ndarray
        An image frame, 2D array of shape (Y, X).
    box : int
        Size of the box to search for local maxima. Should be an odd 
        integer.
        
    Returns
    -------
    y : np.ndarray
        y-coordinates of the local maxima.
    x : np.ndarray
        x-coordinates of the local maxima.
    """
    
    Y, X = frame.shape
    maxima_map = _np.zeros(frame.shape, _np.uint8)
    box_half = int(box / 2)
    box_half_1 = box_half + 1
    for i in range(box_half, Y - box_half_1):
        for j in range(box_half, X - box_half_1):
            local_frame = frame[
                i - box_half : i + box_half + 1,
                j - box_half : j + box_half + 1,
            ]
            flat_max = _np.argmax(local_frame)
            i_local_max = int(flat_max / box)
            j_local_max = int(flat_max % box)
            if (i_local_max == box_half) and (j_local_max == box_half):
                maxima_map[i, j] = 1
    y, x = _np.where(maxima_map)
    return y, x


@_numba.jit(nopython=True, nogil=True, cache=False)
def gradient_at(
    frame: _np.ndarray, 
    y: int, 
    x: int, 
    i: int,
) -> tuple[float, float]:
    """Calculates the gradient at a specific pixel in the frame.
    
    Parameters
    ----------
    frame : np.ndarray
        An image frame, 2D array of shape (Y, X).
    y, x : int
        Coordinates of the pixel where the gradient is calculated.
    i : int
        Index of the pixel in the list of maxima. Not used in this 
        function.
    
    Returns
    -------
    gy : float
        Gradient in the y-direction at the pixel (y, x).
    gx : float
        Gradient in the x-direction at the pixel (y, x).
    """

    gy = frame[y + 1, x] - frame[y - 1, x]
    gx = frame[y, x + 1] - frame[y, x - 1]
    return gy, gx


@_numba.jit(nopython=True, nogil=True, cache=False)
def net_gradient(
    frame: _np.ndarray, 
    y: _np.ndarray, 
    x: _np.ndarray, 
    box: int, 
    uy: _np.ndarray, 
    ux: _np.ndarray,
) -> _np.ndarray:
    """Calculates the net gradient at the identified maxima in the 
    frame.

    Parameters
    ----------
    frame : np.ndarray
        An image frame, 2D array of shape (Y, X).
    y, x : np.ndarray
        Coordinates of the identified maxima in the frame.
    box : int
        Size of the box used for calculating the gradient.
    uy, ux : np.ndarray
        Arrays of shape (box, box) containing the y and x components
        of the gradient, respectively.

    Returns
    -------
    ng : np.ndarray
        Net gradient values at the identified maxima. The shape is
        (len(y),).
    """

    box_half = int(box / 2)
    ng = _np.zeros(len(x), dtype=_np.float32)
    for i, (yi, xi) in enumerate(zip(y, x)):
        for k_index, k in enumerate(range(yi - box_half, yi + box_half + 1)):
            for l_index, m in enumerate(range(xi - box_half, xi + box_half + 1)):
                if not (k == yi and m == xi):
                    gy, gx = gradient_at(frame, k, m, i)
                    ng[i] += gy * uy[k_index, l_index] + gx * ux[k_index, l_index]
    return ng


@_numba.jit(nopython=True, nogil=True, cache=False)
def identify_in_image(
    image: _np.ndarray, 
    minimum_ng: float, 
    box: int,
) -> tuple[_np.ndarray, _np.ndarray, _np.ndarray]:
    """Identifies local maxima in the image and calculates the net
    gradient at those maxima.
    
    Parameters
    ----------
    image : np.ndarray
        An image frame, 2D array of shape (Y, X).
    minimum_ng : float
        Minimum net gradient value to consider a maximum as valid.
    box : int
        Size of the box used for calculating the gradient. Should be
        an odd integer. 
        
    Returns
    -------
    y : np.ndarray
        y-coordinates of the identified maxima.
    x : np.ndarray
        x-coordinates of the identified maxima.
    ng : np.ndarray
        Net gradient values at the identified maxima. The shape is
        (len(y),).
    """

    y, x = local_maxima(image, box)
    box_half = int(box / 2)
    # Now comes basically a meshgrid
    ux = _np.zeros((box, box), dtype=_np.float32)
    uy = _np.zeros((box, box), dtype=_np.float32)
    for i in range(box):
        val = box_half - i
        ux[:, i] = uy[i, :] = val
    unorm = _np.sqrt(ux**2 + uy**2)
    ux /= unorm
    uy /= unorm
    ng = net_gradient(image, y, x, box, uy, ux)
    positives = ng > minimum_ng
    y = y[positives]
    x = x[positives]
    ng = ng[positives]
    return y, x, ng


def identify_in_frame(
    frame: _np.ndarray, 
    minimum_ng: float, 
    box: int, 
    roi: tuple[tuple[int, int], tuple[int, int]] | None = None,
) -> tuple[_np.ndarray, _np.ndarray, _np.ndarray]:
    """Identifies local maxima in a single frame with an optionally
    specified subregion (ROI) and calculates the net gradient at those 
    maxima.
    
    Parameters
    ----------
    frame : np.ndarray
        An image frame, 2D array of shape (Y, X).
    minimum_ng : float
        Minimum net gradient value to consider a maximum as valid.
    box : int
        Size of the box used for calculating the gradient. Should be
        an odd integer.
    roi : tuple, optional
        Region of interest (ROI) defined as a tuple of two tuples,
        where the first tuple contains the start coordinates 
        (y_start, x_start) and the second tuple contains the end 
        coordinates (y_end, x_end). If None, the entire frame is used. 
    
    Returns
    -------
    y : np.ndarray
        y-coordinates of the identified maxima.
    x : np.ndarray
        x-coordinates of the identified maxima.
    net_gradient : np.ndarray
        Net gradient values at the identified maxima. The shape is
        (len(y),).
    """

    if roi is not None:
        frame = frame[roi[0][0] : roi[1][0], roi[0][1] : roi[1][1]]
    image = _np.float32(frame)  # otherwise numba goes crazy
    y, x, net_gradient = identify_in_image(image, minimum_ng, box)
    if roi is not None:
        y += roi[0][0]
        x += roi[0][1]
    return y, x, net_gradient


# def identify_frame(
#     frame: _np.ndarray, 
#     minimum_ng: float, 
#     box: int, 
#     frame_number: int, 
#     roi: tuple[tuple[int, int], tuple[int, int]] | None = None, 
#     resultqueue: _multiprocessing.Queue | None = None
# ) -> _np.recarray:
#     """Identifies local maxima in a single frame and calculates the net
#     gradient at those maxima. Optionally, the results can be put into a
#     multiprocessing queue.

#     Parameters
#     ----------
#     frame : np.ndarray
#         An image frame, 2D array of shape (Y, X).
#     minimum_ng : float
#         Minimum net gradient value to consider a maximum as valid.
#     box : int
#         Size of the box used for calculating the gradient. Should be
#         an odd integer.
#     frame_number : int
#         The index of the frame in the movie sequence.
#     roi : tuple, optional
#         Region of interest (ROI) defined as a tuple of two tuples,
#         where the first tuple contains the start coordinates 
#         (y_start, x_start) and the second tuple contains the end 
#         coordinates (y_end, x_end). If None, the entire frame is used. 
#     resultqueue : multiprocessing.Queue, optional
#         If provided, the results will be put into this queue as a
#         structured numpy array. 

#     Returns
#     -------
#     result : np.recarray
#         A structured numpy array containing the frame number, x and y
#         coordinates of the identified maxima, and their net gradient.   
#     """

#     y, x, net_gradient = identify_in_frame(frame, minimum_ng, box, roi)
#     frame = frame_number * _np.ones(len(x))
#     result = _np.rec.array(
#         (frame, x, y, net_gradient),
#         dtype=[("frame", "i"), ("x", "i"), ("y", "i"), ("net_gradient", "f4")],
#     )
#     if resultqueue is not None:
#         resultqueue.put(result)
#     return result


def identify_by_frame_number(
    movie: _np.ndarray, 
    minimum_ng: float, 
    box: int, 
    frame_number: int, 
    roi: tuple[tuple[int, int], tuple[int, int]] | None = None, 
    lock: _threading.Lock | None = None,
) -> _np.recarray:
    """Identifies local maxima in a specific frame of a movie and
    calculates the net gradient at those maxima. Optionally, a lock can
    be used to ensure thread safety when accessing the movie data.
    
    Parameters
    ----------
    movie : np.ndarray
        A 3D array representing the movie of shape (N, Y, X), where N is
        the number of frames, Y is the height, and X is the width.
    minimum_ng : float
        Minimum net gradient value to consider a maximum as valid.
    box : int
        Size of the box used for calculating the gradient. Should be
        an odd integer.
    frame_number : int
        The index of the frame in the movie sequence to be processed.
    roi : tuple, optional
        Region of interest (ROI) defined as a tuple of two tuples,
        where the first tuple contains the start coordinates 
        (y_start, x_start) and the second tuple contains the end 
        coordinates (y_end, x_end). If None, the entire frame is used.
    lock : threading.Lock, optional
        If provided, this lock will be used to ensure thread safety when
        accessing the movie data. This is useful in a multithreaded
        environment.

    Returns
    -------
    result : np.recarray
        A structured numpy array containing the frame number, x and y
        coordinates of the identified maxima, and their net gradient.
    """

    if lock is not None:
        with lock:
            frame = movie[frame_number]
    else:
        frame = movie[frame_number]
    y, x, net_gradient = identify_in_frame(frame, minimum_ng, box, roi)
    frame = frame_number * _np.ones(len(x))
    return _np.rec.array(
        (frame, x, y, net_gradient),
        dtype=[("frame", "i"), ("x", "i"), ("y", "i"), ("net_gradient", "f4")],
    )


def _identify_worker(
    movie: _np.ndarray, 
    current: list[int], 
    minimum_ng: float,
    box: int,
    roi: tuple[tuple[int, int], tuple[int, int]] | None,
    lock: _threading.Lock | None,
) -> list[_np.recarray]:
    """Worker function for identifying local maxima in a movie. This   
    function is designed to be run in a separate thread and processes
    each frame independently."""

    n_frames = len(movie)
    identifications = []
    while True:
        with lock:
            index = current[0]
            if index == n_frames:
                return identifications
            current[0] += 1
        identifications.append(
            identify_by_frame_number(movie, minimum_ng, box, index, roi, lock)
        )


def identifications_from_futures(
    futures: list[_multiprocessing.pool.Future],
) -> _np.recarray:
    """Collects the results from a list of futures and combines them
    into a single structured numpy array.

    Parameters
    ----------
    futures : list of _multiprocessing.pool.Future's
        A list of futures representing the asynchronous tasks.

    Returns
    -------
    ids : _np.recarray
        A structured numpy array containing the combined results from 
        all futures. Contains fields `frame`, `x`, `y`, and 
        `net_gradient`.
    """

    ids_list_of_lists = [_.result() for _ in futures]
    ids_list = list(_chain(*ids_list_of_lists))
    ids = _np.hstack(ids_list).view(_np.recarray)
    ids.sort(kind="mergesort", order="frame")
    return ids


def identify_async(
    movie: _np.ndarray, 
    minimum_ng: float, 
    box: int, 
    roi: tuple[tuple[int, int], tuple[int, int]] | None = None,
) -> tuple[list[int], list[_multiprocessing.pool.Future]]:
    """Asynchronously (i.e., using multithreading) identifies local 
    maxima in a movie using multiple threads. This function divides the 
    work among a specified number of threads.
    
    Parameters
    ----------
    movie : _np.ndarray
        The input movie data as a 3D numpy array.
    minimum_ng : float
        The minimum net gradient for a spot to be considered.
    box : int
        The size of the box to extract around each spot.
    roi : tuple[tuple[int, int], tuple[int, int]] | None
        The region of interest (ROI) for the analysis.

    Returns
    -------
    current : list[int]
        A list of frame indices representing the current processing 
        state.
    f : list[_multiprocessing.pool.Future]
            A list of futures representing the asynchronous tasks.
    """

    # Use the user settings to define the number of workers that are being used
    settings = _io.load_user_settings()

    # avoid the problem when cpu_utilization is not set
    try:
        cpu_utilization = settings["Localize"]["cpu_utilization"]
    except KeyError:
        cpu_utilization = 0.8

    if isinstance(cpu_utilization, float):
        if cpu_utilization >= 1:
            cpu_utilization = 0.8
    else:
        print("CPU utilization was not set. Setting to 0.8")
        cpu_utilization = 0.8
    settings["Localize"]["cpu_utilization"] = cpu_utilization
    _io.save_user_settings(settings)

    n_workers = min(
        60, max(1, int(cpu_utilization * _multiprocessing.cpu_count()))
    ) # Python crashes when using >64 cores

    lock = _threading.Lock()
    current = [0]
    executor = _ThreadPoolExecutor(n_workers)
    f = [
        executor.submit(_identify_worker, movie, current, minimum_ng, box, roi, lock)
        for _ in range(n_workers)
    ]
    executor.shutdown(wait=False)
    return current, f


def identify(
    movie: _np.ndarray, 
    minimum_ng: float, 
    box: int, 
    threaded: bool = True,
) -> _np.recarray:
    """Identifies local maxima in a movie and calculates the net
    gradient at those maxima. This function can run in a threaded or
    non-threaded mode.

    Parameters
    ----------
    movie : _np.ndarray
        The input movie data as a 3D numpy array.
    minimum_ng : float
        The minimum net gradient for a spot to be considered.
    box : int
        The size of the box to extract around each spot.
    threaded : bool
        Whether to use threading for the identification process.

    Returns
    -------
    ids : _np.recarray
        A structured numpy array containing the identified spots. 
        Contains fields `frame`, `x`, `y`, and `net_gradient`.
    """

    if threaded:
        current, futures = identify_async(movie, minimum_ng, box)
        identifications = [_.result() for _ in futures]
        identifications = [_np.hstack(_) for _ in identifications]
    else:
        identifications = [
            identify_by_frame_number(movie, minimum_ng, box, i)
            for i in range(len(movie))
        ]
    ids = _np.hstack(identifications).view(_np.recarray)
    return ids


@_numba.jit(nopython=True, cache=False)
def _cut_spots_numba(
    movie: _np.ndarray, 
    ids_frame: _np.ndarray, 
    ids_x: _np.ndarray, 
    ids_y: _np.ndarray, 
    box: int,
) -> _np.ndarray:
    """Extracts the spots out of a movie using Numba for performance."""
    
    n_spots = len(ids_x)
    r = int(box / 2)
    spots = _np.zeros((n_spots, box, box), dtype=movie.dtype)
    for id, (frame, xc, yc) in enumerate(zip(ids_frame, ids_x, ids_y)):
        spots[id] = movie[frame, yc - r : yc + r + 1, xc - r : xc + r + 1]
    return spots


@_numba.jit(nopython=True, cache=False)
def _cut_spots_frame(
    frame: _np.ndarray, 
    frame_number: int, 
    ids_frame: _np.ndarray, 
    ids_x: _np.ndarray, 
    ids_y: _np.ndarray, 
    r: int, 
    start: int, 
    N: int, 
    spots: _np.ndarray,
) -> int:
    """Extracts spots from a movie frame."""

    for j in range(start, N):
        if ids_frame[j] > frame_number:
            break
        if ids_frame[j] < frame_number:
            break
        yc = ids_y[j]
        xc = ids_x[j]
        spots[j] = frame[yc - r : yc + r + 1, xc - r : xc + r + 1]
    return j


@_numba.jit(nopython=True, cache=False)
def _cut_spots_daskmov(movie, l_mov, ids_frame, ids_x, ids_y, box, spots):
    """Extracts the spots out of a movie frame by frame.

    Parameters
    ----------
    movie : _np.ndarray
        The input movie data as a 3D numpy array.
    l_mov : _np.ndarray
        Length of the movie, a 1D array with a single element.
    ids_frame, ids_x, ids_y : _np.ndarray
        1D arrays containing spot positions in the image data.
    box : int
        Size of the box to cut out around each spot. Should be an odd
        integer.
    spots : _np.ndarray
        3D array to store the cut spots, with shape (k, box, box), 
        where k is the number of spots identified.
    
    Returns
    -------
    spots : _np.ndarray
        3D array with extracted spots of shape (k, box, box), where k is 
        the number of spots identified.
    """
    r = int(box / 2)
    N = len(ids_frame)
    start = 0
    for frame_number in range(l_mov[0]):
        frame = movie[frame_number, :, :]
        start = _cut_spots_frame(
            frame,
            frame_number,
            ids_frame,
            ids_x,
            ids_y,
            r,
            start,
            N,
            spots,
        )
    return spots


def _cut_spots_framebyframe(movie, ids_frame, ids_x, ids_y, box, spots):
    """Extracts the spots out of a movie frame by frame.

    Parameters
    ----------
    movie : _np.ndarray
        The input movie data as a 3D numpy array.
    ids_frame, ids_x, ids_y : _np.ndarray
        1D arrays containing spot positions in the image data.
    box : int
        Size of the box to cut out around each spot. Should be an odd
        integer.
    spots : _np.ndarray
        3D array to store the cut spots, with shape (k, box, box), 
        where k is the number of spots identified.
    
    Returns
    -------
    spots : _np.ndarray
        3D array with extracted spots of shape (k, box, box), where k is 
        the number of spots identified.
    """

    r = int(box / 2)
    N = len(ids_frame)
    start = 0
    for frame_number, frame in enumerate(movie):
        start = _cut_spots_frame(
            frame,
            frame_number,
            ids_frame,
            ids_x,
            ids_y,
            r,
            start,
            N,
            spots,
        )
    return spots


def _cut_spots(movie: _np.ndarray, ids: _np.ndarray, box: int) -> _np.ndarray:
    """Cuts out spots from a movie based on the identified positions."""

    N = len(ids.frame)
    if isinstance(movie, _np.ndarray):
        return _cut_spots_numba(movie, ids.frame, ids.x, ids.y, box)
    elif isinstance(movie, _io.ND2Movie) and movie.use_dask:
    # elif movie.use_dask:
        """ Assumes that identifications are in order of frames! """
        spots = _np.zeros((N, box, box), dtype=movie.dtype)
        spots = _da.apply_gufunc(
            _cut_spots_daskmov,
            '(p,n,m),(b),(k),(k),(k),(),(k,l,l)->(k,l,l)',
            movie.data, _np.array([len(movie)]), ids.frame, ids.x, ids.y, box, spots,
            output_dtypes=[movie.dtype], allow_rechunk=True).compute()
        return spots
    else:
        """Assumes that identifications are in order of frames!"""
        
        r = int(box / 2)
        N = len(ids.frame)
        spots = _np.zeros((N, box, box), dtype=movie.dtype)
        spots = _cut_spots_framebyframe(
            movie, ids.frame, ids.x, ids.y, box, spots)
        return spots


def _to_photons(spots: _np.ndarray, camera_info: dict) -> _np.ndarray:
    """Converts the cut spots to photon counts based on camera 
    information."""

    spots = _np.float32(spots)
    baseline = camera_info["Baseline"]
    sensitivity = camera_info["Sensitivity"]
    gain = camera_info["Gain"]
    # since v0.6.0: remove quantum efficiency to better reflect precision
    # qe = camera_info["Qe"]
    return (spots - baseline) * sensitivity / (gain)


def get_spots(
    movie: _np.ndarray, 
    identifications: _np.recarray, 
    box: int, 
    camera_info: dict,
) -> _np.ndarray:
    """Extracts spots from a movie based on the identified positions and
    converts camera signal to photon counts.
    
    Parameters
    ----------
    movie : _np.ndarray
        The input movie data as a 3D numpy array.
    identifications : _np.recarray
        A structured numpy array containing the identified spots.
        Contains fields `frame`, `x`, `y`, and `net_gradient`.
    box : int
        Size of the box to cut out around each spot. Should be an odd
        integer.
    camera_info : dict
        A dictionary containing camera information such as
        `Baseline`, `Sensitivity`, and `Gain`.
    
    Returns
    -------
    spots : _np.ndarray
        A 3D numpy array containing the extracted spots, with shape
        (k, box, box), where k is the number of spots identified.
    """

    spots = _cut_spots(movie, identifications, box)
    return _to_photons(spots, camera_info)


def fit(
    movie: _np.ndarray,
    camera_info: dict,
    identifications: _np.recarray,
    box: int,
    eps: float = 0.001,
    max_it: int = 100,
    method: Literal["sigma", "sigmaxy"] = "sigma",
) -> _np.recarray:
    """Fits Gaussians using Maximum Likelihood Estimation (MLE) to the 
    identified spots in a movie to localize fluorescent molecules. See
    Smith, et al. Nature Methods, 2010. DOI: 10.1038/nmeth.1449.
    
    Parameters
    ----------
    movie : _np.ndarray
        The input movie data as a 3D numpy array.
    camera_info : dict
        A dictionary containing camera information such as
        `Baseline`, `Sensitivity`, and `Gain`.
    identifications : _np.recarray
        A structured numpy array containing the identified spots.
        Contains fields `frame`, `x`, `y`, and `net_gradient`.
    box : int
        Size of the box to cut out around each spot. Should be an odd
        integer.
    eps : float, optional
        The convergence criterion for the fitting algorithm. Default is
        0.001.
    max_it : int, optional
        The maximum number of iterations for the fitting algorithm. 
        Default is 100.
    method : Literal["sigma", "sigmaxy"], optional
        The method used for fitting (impose same sigma in x and y or 
        not, respectively). Default is "sigma".

    Returns
    -------
    locs : _np.recarray
        A structured numpy array containing the localized spots. The
        fields include `frame`, `x`, `y`, `photons`, `sx`, `sy`, `bg`, 
        `lpx`, `lpy`, `net_gradient`, `likelihood`, and `iterations`.
    """
    
    spots = get_spots(movie, identifications, box, camera_info)
    theta, CRLBs, likelihoods, iterations = _gaussmle.gaussmle(
        spots, eps, max_it, method=method
    )
    return locs_from_fits(identifications, theta, CRLBs, likelihoods, iterations, box)


def fit_async(
    movie: _np.ndarray,
    camera_info: dict,
    identifications: _np.recarray,
    box: int,
    eps: float = 0.001,
    max_it: int = 100,
    method: Literal["sigma", "sigmaxy"] = "sigmaxy",
) -> tuple[int, _np.ndarray, _np.ndarray, _np.ndarray, _np.ndarray]:
    """Asynchronously fits Gaussians using Maximum Likelihood Estimation
    (MLE) to the identified spots in a movie to localize fluorescent
    molecules. This function is designed to run in a separate thread or
    process. See Smith, et al. Nature Methods, 2010. 
    DOI: 10.1038/nmeth.1449.

    Parameters
    ----------
    movie : _np.ndarray
        The input movie data as a 3D numpy array.
    camera_info : dict
        A dictionary containing camera information such as
        `Baseline`, `Sensitivity`, and `Gain`.
    identifications : _np.recarray
        A structured numpy array containing the identified spots.
        Contains fields `frame`, `x`, `y`, and `net_gradient`.
    box : int
        Size of the box to cut out around each spot. Should be an odd
        integer.
    eps : float, optional       
        The convergence criterion for the fitting algorithm. Default is
        0.001.
    max_it : int, optional  
        The maximum number of iterations for the fitting algorithm.
        Default is 100.
    method : Literal["sigma", "sigmaxy"], optional
        The method used for fitting (impose same sigma in x and y or 
        not, respectively). Default is "sigmaxy".

    Returns
    -------
    current : int
        Index of the currently processed spot.
    thetas : _np.ndarray
        The fitted Gaussian parameters for each spot (x, y positions, 
        photon counts, background, single-emitter image size in x and 
        y).
    CRLBs : _np.ndarray
        The Cramer-Rao Lower Bounds for each fitted parameter.
    likelihoods : _np.ndarray
        The log-likelihoods of the fitted models.
    iterations : _np.ndarray
        The number of iterations taken to converge for each spot.
    """

    spots = get_spots(movie, identifications, box, camera_info)
    return _gaussmle.gaussmle_async(spots, eps, max_it, method=method)


def locs_from_fits(
    identifications: _np.recarray, 
    theta: _np.ndarray, 
    CRLBs: _np.ndarray, 
    likelihoods: _np.ndarray, 
    iterations: _np.ndarray, 
    box: int,
) -> _np.recarray:
    """Converts the resulting localizations from the list of Futures 
    into a structured array.

    Parameters
    ----------
    identifications : _np.recarray
        A structured numpy array containing the identified spots.
        Contains fields `frame`, `x`, `y`, and `net_gradient`.
    theta : _np.ndarray
        The fitted Gaussian parameters for each spot (x, y positions, 
        photon counts, background, single-emitter image size in x and 
        y).
    CRLBs : _np.ndarray
        The Cramer-Rao Lower Bounds for each fitted parameter.
    likelihoods : _np.ndarray
        The log-likelihoods of the fitted models.
    iterations : _np.ndarray
        The number of iterations taken to converge for each spot.       
    box : int
        Size of the box used for fitting. Should be an odd integer. 
    
    Returns
    -------
    locs : _np.recarray
        A structured numpy array containing the localized spots. The
        fields include `frame`, `x`, `y`, `photons`, `sx`, `sy`, `bg`, 
        `lpx`, `lpy`, `net_gradient`, `likelihood`, and `iterations`.
    """

    box_offset = int(box / 2)
    y = theta[:, 0] + identifications.y - box_offset
    x = theta[:, 1] + identifications.x - box_offset
    lpy = _np.sqrt(CRLBs[:, 0])
    lpx = _np.sqrt(CRLBs[:, 1])
    locs = _np.rec.array(
        (
            identifications.frame,
            x,
            y,
            theta[:, 2],
            theta[:, 5],
            theta[:, 4],
            theta[:, 3],
            lpx,
            lpy,
            identifications.net_gradient,
            likelihoods,
            iterations,
        ),
        dtype=LOCS_DTYPE,
    )
    locs.sort(kind="mergesort", order="frame")
    return locs


def localize(
    movie: _np.ndarray, 
    camera_info: dict, 
    parameters: dict,
) -> _np.recarray:
    """Localizes (i.e., identifies and fits) spots in a movie using
    the specified parameters. This function combines the identification
    and fitting steps into a single process.
    
    Parameters
    ----------
    movie : _np.ndarray
        The input movie data as a 3D numpy array.
    camera_info : dict
        A dictionary containing camera information such as
        `Baseline`, `Sensitivity`, and `Gain`.
    parameters : dict
        A dictionary containing localization parameters, including:
        - `Min. Net Gradient`: Minimum net gradient for spot
          identification.
        - `Box Size`: Size of the box to cut out around each spot.

    Returns
    -------
    locs : _np.recarray
        A structured numpy array containing the localized spots.
        The fields include `frame`, `x`, `y`, `photons`, `sx`, `sy`,
        `bg`, `lpx`, `lpy`, `net_gradient`, `likelihood`, and 
        `iterations`.
    """

    identifications = identify(
        movie,
        parameters["Min. Net Gradient"],
        parameters["Box Size"],
    )
    locs = fit(movie, camera_info, identifications, parameters["Box Size"])
    return locs


def check_nena(
    locs: _np.recarray, 
    info: None, 
    callback: Callable[[int], None] = None,
) -> float:
    """Calculates the NeNA (experimental localization precision) from
    localizations.
    
    Parameters
    ----------
    locs : _np.recarray
        A structured numpy array containing the localized spots.
    info : None
        Not used.
    callback : Callable[[int], None], optional
        A callback function that can be used to report progress. It
        should accept an integer argument representing the current
        step or frame number. Default is None.
    
    Returns
    -------
    nena_px : float
        The NeNA value in pixels, representing the experimental
        localization precision.
    """

    # Nena
    print('Calculating NeNA.. ', end ='')
    locs = locs[0:MAX_LOCS]
    try:
        result, best_result = _postprocess.nena(locs, None, callback=callback)
        nena_px = best_result
    except Exception as e:
        print(e)
        nena_px = float("nan")

    print(f"{nena_px:.2f} px.")

    return nena_px


def check_kinetics(locs: _np.recarray, info: list[dict]) -> float:
    """Calculates the mean length of binding events from localizations.

    Parameters
    ----------
    locs : _np.recarray
        A structured numpy array containing the localized spots.
    info : list[dict]
        A list of dictionaries containing metadata about the movie.
    
    Returns
    -------
    len_mean : float
        The mean length of binding events in frames.    
    """

    print("Linking.. ", end ='')
    locs = locs[0:MAX_LOCS]
    locs = _postprocess.link(locs, info=info)
    len_mean = locs.len.mean()
    print(f"Mean length {len_mean:.2f} frames.")

    return len_mean


def check_drift(
    locs: _np.recarray, 
    info: list[dict], 
    callback: Callable[[int], None] = None,
) -> tuple[float, float]:
    """Estimates the drift of localizations in x and y directions.
    
    Parameters
    ----------
    locs : _np.recarray
        A structured numpy array containing the localized spots.
    info : list[dict]
        A list of dictionaries containing metadata about the movie.
    callback : Callable[[int], None], optional
        A callback function that can be used to report progress. It 
        should accept an integer argument representing the current
        step or frame number. Default is None.

    Returns
    -------
    drift_x : float
        The estimated drift in the x direction.
    drift_y : float
        The estimated drift in the y direction.
    """

    steps = int(len(locs) // (MAX_LOCS))
    steps = max(1, steps)

    locs = locs[::steps]

    n_frames = info[0]["Frames"]

    segmentation = max(1, int(n_frames // 10))

    print(f"Estimating drift with segmentation {segmentation}")
    drift, locs = _postprocess.undrift(
        locs,
        info,
        segmentation,
        display=False,
        rcc_callback=callback,
    )

    drift_x = float(drift["x"].mean())
    drift_y = float(drift["y"].mean())

    print(f"Drift is X: {drift_x:.2f}, Y: {drift_y:.2f}.")

    return (drift_x, drift_y)


def get_file_summary(
    file: str, 
    file_hdf: str, 
    drift: tuple[float, float] | None = None, 
    len_mean: float | None = None, 
    nena: float | None = None,
) -> dict:
    """Generates a summary of the localization file, including metadata
    and statistics about the localizations.

    Parameters
    ----------
    file : str
        The path to the localization file (HDF5 format).
    file_hdf : str
        The path to the HDF5 file containing localizations.
    drift : tuple[float, float] | None, optional
        A tuple containing the drift in x and y directions. If None,
        the drift will be calculated from the localizations.
    len_mean : float | None, optional
        The mean length of binding events in frames. If None, it will
        be calculated from the localizations.
    nena : float | None, optional
        The NeNA value in pixels. If None, it will be calculated from
        the localizations.

    Returns
    -------
    summary : dict
        A dictionary containing the summary of the localization file,
        including metadata and statistics about the localizations.
    """

    if file_hdf is None:
        base, ext = os.path.splitext(file)
        file_hdf = base + "_locs.hdf5"

    locs, info = _io.load_locs(file_hdf)

    summary = {}

    for col in MEAN_COLS:
        try:
            summary[col + "_mean"] = locs[col].mean()
            summary[col + "_std"] = locs[col].std()
        except ValueError:
            summary[col + "_mean"] = float("nan")
            summary[col + "_std"] = float("nan")

    for col in SET_COLS:
        col_ = col.lower()
        for inf in info:
            if col in inf:
                summary[col_] = inf[col]

    for col in SET_COLS:
        col_ = col.lower()
        if col_ not in summary:
            summary[col_] = float("nan")

    if nena is None:
        summary["nena_px"] = check_nena(locs, info)
    else:
        summary["nena_px"] = nena

    if len_mean is None:
        len_mean = check_kinetics(locs, info)
    else:
        len_mean = len_mean

    if drift is None:
        drift_x, drift_y = check_drift(locs, info)
    else:
        drift_x, drift_y = drift

    summary["len_mean"] = len_mean
    summary["n_locs"] = len(locs)
    summary["locs_frame"] = len(locs) / summary["frames"]

    summary["drift_x"] = drift_x
    summary["drift_y"] = drift_y

    summary["nena_nm"] = summary["nena_px"] * summary["pixelsize"]

    summary["filename"] = os.path.normpath(file)
    summary["filename_hdf"] = file_hdf
    summary["file_created"] = datetime.fromtimestamp(os.path.getmtime(file))
    summary["entry_created"] = datetime.now()

    return summary


def _db_filename() -> str:
    """Returns the path to the SQLite database file used for storing
    localization summaries. The database is stored in the user's home
    directory under the `.picasso` folder."""

    home = os.path.expanduser("~")
    picasso_dir = os.path.join(home, ".picasso")
    os.makedirs(picasso_dir, exist_ok=True)
    return os.path.abspath(os.path.join(picasso_dir, "app_0410.db"))


def save_file_summary(summary: dict) -> None:
    """Saves the summary of a localization file to a SQLite database."""

    engine = create_engine("sqlite:///" + _db_filename(), echo=False)
    s = pd.Series(summary, index=summary.keys()).to_frame().T
    s.to_sql("files", con=engine, if_exists="append", index=False)


def add_file_to_db(
    file: str, 
    file_hdf: str, 
    drift: tuple[float, float] | None = None, 
    len_mean: float | None = None, 
    nena: float | None = None
) -> None:
    """Adds a localization file summary to the SQLite database."""
    
    summary = get_file_summary(file, file_hdf, drift, len_mean, nena)
    save_file_summary(summary)